"""
Fluid Block
===========

This file implements an extension of the  basic SysML Block to specify
interactions along a continuous fluid system.

The model is intended for steady-state incompressible flow and is quite rough.
"""
from __future__ import annotations

from typing import List
import numpy as np

import cetpy.Modules.SysML as SML
from cetpy.Modules.SysML import value_property, DeterminationTest
from cetpy.Modules.Fluid import FluidSkeleton
from cetpy.Modules.FluidBlock import FluidPort

from cetpy.Modules.Solver import Solver


def apply_transfer(block: FluidBlock,
                   flow_property: SML.FlowProperty,
                   simple_only: bool = False) -> None:
    name = flow_property.__getattribute__('_name')
    inlet = block.inlet_no_solve
    outlet = block.outlet_no_solve
    direction = flow_property.get_direction(block.inlet_no_solve)

    if direction == 'downstream':
        source = inlet
        target = outlet
        factor = 1
    else:
        source = outlet
        target = inlet
        factor = -1

    flow_property.__set_converging_value__(
        target, (flow_property.__get__(source)
                 + factor * block.__getattribute__(f'_d{name}_stored')))
    if not simple_only:
        flow_property.__set_converging_value__(
            target, (flow_property.__get__(source)
                     + factor * block.__getattribute__('d' + name)))


class FluidSolver(Solver):
    """Specification of the Solver class to solve a continuous fluid system."""

    __slots__ = ['parents', '_flow_properties']

    def __init__(self, parent: FluidBlock, tolerance: float = None):
        self._flow_properties = None
        super().__init__(parent, tolerance)
        self.parents = [parent]

    # region Input Properties
    @property
    def flow_properties(self) -> List[SML.FlowProperty]:
        if self._flow_properties is None:
            try:
                port = self.parent.outlet_no_solve
            except AttributeError:
                port = self.parent.inlet_no_solve
            self._flow_properties = port.__flow_properties__
        return self._flow_properties

    @property
    def ordered_solver_lists(self) -> List[List[SML.Block]]:
        """Lists of connected fluid blocks for each flow property in the
        direction of the solver."""
        try:
            blocks = self.parent.outlet_no_solve.flow_system_list
        except AttributeError:
            blocks = self.parent.inlet_no_solve.flow_system_list

        flow_properties = self.flow_properties
        ordered_lists = []
        for fp in flow_properties:
            fp_direction = fp.get_direction(blocks[-1].inlet_no_solve)
            if fp_direction == 'upstream':
                ordered_lists += [blocks.copy()]
                ordered_lists[-1].reverse()
            else:
                ordered_lists += [blocks.copy()]
        return ordered_lists
    # endregion

    def __solve_simple_step__(self) -> None:
        """Propagate boundary conditions through fluid system without
        running any element solvers."""
        flow_properties = self.flow_properties
        ordered_lists = self.ordered_solver_lists

        calculating = self.calculating
        self._calculating = True
        [[apply_transfer(b, fp, simple_only=True) for b in l]
         for fp, l in zip(flow_properties, ordered_lists)]
        self._calculating = calculating

    def _solve_function(self) -> None:
        """Solve the fluid system transfer functions until convergence."""
        self.__solve_simple_step__()

        flow_properties = self.flow_properties
        ordered_lists = self.ordered_solver_lists

        try:
            ports = self.parent.outlet_no_solve.flow_system_port_list
        except AttributeError:
            ports = self.parent.inlet_no_solve.flow_system_port_list

        values_last = np.asarray(
            [[fp.__get__(p) for p in ports] for fp in flow_properties])

        iteration = 0
        tolerance = self.tolerance
        residuals = np.ones((len(flow_properties), len(ports)))

        while np.any(residuals > tolerance):
            iteration += 1
            if iteration > 25:
                raise ValueError(f"Fluid System solver did not converge in"
                                 f"{iteration - 1} iterations.")

            [[apply_transfer(b, fp) for b in l]
             for fp, l in zip(flow_properties, ordered_lists)]

            values = np.asarray(
                [[fp.__get__(p) for p in ports] for fp in flow_properties])

            residuals = np.abs(values - values_last) / values
            values_last = values


class FluidBlock(SML.Block):
    """Fluid Block element."""

    _reset_dict = SML.Block._reset_dict
    _reset_dict.update({'_dp_recalculate': True, '_dt_recalculate': True,
                        '_dmdot_recalculate': True})
    _hard_reset_dict = SML.Block._hard_reset_dict
    _hard_reset_dict.update({'_dp_stored': True, '_dt_stored': True,
                             '_dmdot_stored': True})
    dp_fixed = SML.ValueProperty()
    dt_fixed = SML.ValueProperty()
    dmdot_fixed = SML.ValueProperty()

    __init_parameters__ = SML.Block.__init_parameters__ + [
        'dp_fixed', 'dt_fixed', 'dmdot_fixed', 'area', 'hydraulic_diameter'
    ]

    def __init__(self, name: str, abbreviation: str = None,
                 parent: SML.Block = None,
                 tolerance: float = None,
                 upstream: FluidBlock = None,
                 downstream: FluidBlock = None,
                 fluid: FluidSkeleton | str = None,
                 dp_fixed: float = 0,
                 dt_fixed: float = 0,
                 dmdot_fixed: float = 0,
                 inlet_pressure: float = None,
                 outlet_pressure: float = None,
                 inlet_temperature: float = None,
                 outlet_temperature: float = None,
                 inlet_mass_flow: float = None,
                 outlet_mass_flow: float = None,
                 pressure_solver_direction: str = None,
                 temperature_solver_direction: str = None,
                 mass_flow_solver_direction: str = None,
                 **kwargs) -> None:
        self._dp_recalculate = True
        self._dt_recalculate = True
        self._dmdot_recalculate = True
        self._dp_stored = 0
        self._dt_stored = 0
        self._dmdot_stored = 0

        super().__init__(name=name, abbreviation=abbreviation, parent=parent,
                         tolerance=tolerance,
                         dp_fixed=dp_fixed, dt_fixed=dt_fixed,
                         dmdot_fixed=dmdot_fixed,
                         **kwargs)

        # region Port References
        # To enable detection of inlet as inlet, otherwise the direction is
        # misidentified.
        self._outlet = FluidPort()
        self._inlet = FluidPort(
            downstream=self, flow_item=fluid, pressure=inlet_pressure,
            temperature=inlet_temperature, mass_flow=inlet_mass_flow)
        self._outlet = FluidPort(
            upstream=self, flow_item=fluid, pressure=outlet_pressure,
            temperature=outlet_temperature, mass_flow=outlet_mass_flow)
        self.ports += [self._inlet, self._outlet]
        self.fluid_solver = FluidSolver(parent=self)
        # endregion

        # region Verify Input Validity
        if self._inlet.upstream is not None and (
                inlet_mass_flow is not None or
                inlet_pressure is not None or
                inlet_temperature is not None):
            raise ValueError("Incompatible input, an upstream element is set "
                             "but custom inlet conditions are defined.")
        if self._outlet.downstream is not None and (
                outlet_mass_flow is not None or
                outlet_pressure is not None or
                outlet_temperature is not None):
            raise ValueError("Incompatible input, a downstream element is set "
                             "but custom outlet conditions are defined.")
        for prop, inlet, outlet, direction in zip(
                ['Pressure', 'Temperature', 'Mass flow'],
                [inlet_pressure, inlet_temperature, inlet_mass_flow],
                [outlet_pressure, outlet_temperature, outlet_mass_flow],
                [pressure_solver_direction, temperature_solver_direction,
                 mass_flow_solver_direction]):
            if direction == 'downstream' and outlet is not None:
                raise ValueError(f"{prop} input mismatch, the property is "
                                 f"solved from the inlet, but a custom "
                                 f"outlet value is defined.")
            if direction == 'upstream' and inlet is not None:
                raise ValueError(f"{prop} input mismatch, the property is "
                                 f"solved from the outlet, but a custom "
                                 f"inlet value is defined.")
        # endregion

        # region Fluid References

        # Do this after verifying the input is valid so as not to corrupt
        # the up- or downstream references.
        self.upstream = upstream
        self.downstream = downstream
        # endregion

    # region Transfer Functions
    def _dp_solve(self) -> float:
        """Pressure difference across the fluid element [Pa].

        This is the private calculate function.

        See Also
        --------
        FluidBlock.dp: public function with necessity check
        """
        return self.dp_fixed

    def _dt_solve(self) -> float:
        """Temperature difference across the fluid element [K].

        This is the private calculate function.

        See Also
        --------
        FluidBlock.dt: public function with necessity check
        """
        return self.dt_fixed

    def _dmdot_solve(self) -> float:
        """Mass flow difference across the fluid element [kg/s].

        This is the private calculate function.

        See Also
        --------
        FluidBlock.dmdot: public function with necessity check
        """
        return self.dmdot_fixed

    @value_property()
    def dp(self) -> float:
        r"""Difference in pressure from inlet to outlet [Pa].

        ..math::
            \Delta p = p_{out} - p_{in}
        """
        if self._dp_recalculate:
            self._dp_stored = self._dp_solve()
            self._dp_recalculate = False
        return self._dp_stored

    @value_property()
    def dt(self) -> float:
        r"""Difference in temperature from inlet to outlet [K].

        ..math::
            \Delta T = T_{out} - T_{in}
        """
        if self._dt_recalculate:
            self._dt_stored = self._dt_solve()
            self._dt_recalculate = False
        return self._dt_stored

    @value_property()
    def dmdot(self) -> float:
        r"""Difference in mass flow from inlet to outlet [kg/s].

        ..math::
            \Delta \dot{m} = \dot{m}_{out} - \dot{m}_{in}
        """
        if self._dmdot_recalculate:
            self._dmdot_stored = self._dmdot_solve()
            self._dmdot_recalculate = False
        return self._dmdot_stored
    # endregion

    # region System References and Ports
    @property
    def upstream(self) -> SML.Block | None:
        """Fluid block upstream element."""
        return self._inlet.upstream

    @upstream.setter
    def upstream(self, val: SML.Block | None):
        self._inlet.upstream = val

    @property
    def downstream(self) -> SML.Block | None:
        """Fluid block downstream element."""
        return self._outlet.downstream

    @downstream.setter
    def downstream(self, val: SML.Block | None):
        self._outlet.downstream = val

    @property
    def inlet(self) -> FluidPort:
        """Fluid block inlet port with solved properties.

        See Also
        --------
        FluidBlock.inlet_no_solve: Useful alternative to set new properties
                                   without first calling the solver.
        """
        self.fluid_solver.solve()
        return self._inlet

    @property
    def inlet_no_solve(self) -> FluidPort:
        """Fluid block inlet port without solver properties.

        This should only be used for setting new parameters without calling
        the solver. Do not trust port value properties when accessing through
        this function.

        See Also
        --------
        FluidBlock.inlet
        """
        return self._inlet

    @property
    def outlet(self) -> FluidPort:
        """Fluid block outlet port with solved properties.

        See Also
        --------
        FluidBlock.outlet_no_solve: Useful alternative to set new properties
                                    without first calling the solver.
        """
        self.fluid_solver.solve()
        return self._outlet

    @property
    def outlet_no_solve(self) -> FluidPort:
        """Fluid block outlet port without solver properties.

        This should only be used for setting new parameters without calling
        the solver. Do not trust port value properties when accessing through
        this function.

        See Also
        --------
        FluidBlock.inlet
        """
        return self._outlet
    # endregion

    # region Input Properties
    @value_property(determination_test=DeterminationTest())
    def area(self) -> float:
        """Fluid element characteristic flow area [m^2]."""
        return np.pi * (self.hydraulic_diameter / 2) ** 2

    @value_property(determination_test=area.determination_test)
    def hydraulic_diameter(self) -> float:
        """Fluid block characteristic hydraulic diameter [m]."""
        return np.sqrt(self.area / np.pi) * 2
    # endregion



"""
Module Manager
==============

This file defines function to aggregate a list of modules used in cetpy and
extension modules. It pulls from the cetpy.Configuration module_location
list and aggregates all modules deriving from the base SysML Block class.
"""
import sys
import pkgutil
import os.path
from types import ModuleType
from typing import Dict


import cetpy
import cetpy.Configuration
from cetpy.Modules.SysML import Block


def get_module(name: str) -> cetpy.Modules.SysML.Block:
    """Find and return a module based on its base name from the blocks in
    the module locations of the cetpy configuration."""
    return cetpy.Configuration.module_dict[name]


def generate_block_class_list() -> Dict[str, Block]:
    """Return dictionary of all Block derived classes in across cetpy and
    extension modules."""
    blocks = {}
    for location in cetpy.Configuration.module_locations:
        location = os.path.abspath(location)
        # The issue here is that the location, while on the OS path, can be
        # n-levels deep in this path. So we must strip out the overlap
        # between the path entry and the module locations. Finally, we must
        # swap out the directory separators for '.' package separators.
        overlap = [p for p in sys.path if p in location][-1]
        location_new = location.replace(overlap, '').replace(
            os.path.sep, '.')[1:]

        module_instance = get_submodule_instance(location_new)

        # Verify that the correct module was loaded, also important for
        # security.
        if module_instance.__path__[0] != location:
            raise ValueError(f"The path match failed on the module locations. "
                             f"The desired import path was {location}, "
                             f"the attempted match resulted in "
                             f"{module_instance.__path__[0]}.")

        blocks.update(get_module_recursive_classes(module_instance))
    return blocks


def get_module_recursive_classes(ref_module: ModuleType) -> Dict[str, Block]:
    """Return dictionary of name and type of all classes in a module.
    Function is recursive also pulling from submodules."""
    blocks = {}
    blocks.update(get_module_classes(ref_module))
    for submodule in get_submodules(ref_module).values():
        blocks.update(get_module_recursive_classes(submodule))
    return blocks


def get_module_classes(ref_module: ModuleType) -> Dict[str, Block]:
    """Return dictionary of name and type of all classes in a module."""
    blocks = {}
    module_dict = ref_module.__dict__
    for c in module_dict:
        if isinstance(module_dict[c], type) and Block in module_dict[c].mro():
            blocks.update({c: module_dict[c]})
    return blocks


def get_submodules(ref_module: ModuleType) -> Dict[str, ModuleType]:
    """Return dictionary of name and Module Instance of all submodules of a
    module."""
    mod_list = list([
        i for i in pkgutil.iter_modules(ref_module.__path__,
                                        prefix=ref_module.__name__ + '.')
        if i.ispkg])
    modules = {}
    for submodule in mod_list:
        submodule_instance = get_submodule_instance(submodule.name)
        modules.update({submodule.name: submodule_instance})

    return modules


def get_submodule_instance(name: str) -> ModuleType:
    """Return instance of a submodule."""
    submodule = __import__(name)
    if '.' not in name:
        return submodule
    name_split = name.split('.')[1:]
    for n in name_split:
        if n in submodule.__dict__.keys():
            submodule = submodule.__dict__[n]
    return submodule



import cetpy.Configuration
from cetpy.Configuration import Session

active_session = Session(None)
sessions = []
cetpy.Configuration.refresh_module_dict()


def refresh() -> None:
    """Refresh all cetpy Sessions and the module list."""
    active_session.refresh()
    for sess in sessions:
        sess.refresh()



"""
Input Validation
================

Provides functions related to verifying types, values, and ranges of input
values.
"""
from typing import Any, Tuple, List, Iterable
import numpy as np


def validate_input(value: Any, permissible_types: None | type | List = None,
                   permissible: None | List | Tuple = None,
                   property_name: str = None):
    """Test the input value against the permissible type and value lists
    and correct any minor deviations.

    Properties
    ----------
    value
        The input value which should be verified against the provided criteria.
    permissible_types: optional, default = None
        List of permissible python types which the value should have. Set to
        None to disable. A single type is a permissible input. small
        deviations are corrected automatically. Notably, bool, int,
        and float, values are converted to target types if they are any one of
        bool, int, float, str. Additionally, single values can be converted
        to 1d np arrays, or general iterable types to numpy arrays.
    permissible: optional, default = None
        Description of range testing of the value or individual values. Pass a
        list to describe a set of permissible specific values, or a tuple of
        length 2 to specify a minium and maximum value. The limits are
        inclusive. None is equivalent to -np.inf, and np.inf when specified
        in the first or second index. Pass None to disable.
    property_name: optional, default = None
        An optional property name of the input property, this allows
        provision of clearer error messages.

    Returns
    -------
    Any
        The input value, converted to a permissible type and within a
        permissible range.

    Raises
    ------
    ValueError
        If either the type cannot be matched or automatically corrected,
        or the value is outside the permissible limits or not in the
        permissible set.

    Examples
    --------
    >>> validate_input(1, float, (0, 1))
    1.0
    >>> validate_input(3., bool)
    True
    >>> validate_input(True, float, (0, None))
    1.0
    >>> validate_input(8.756, str)
    '8.756'
    >>> validate_input(42.24, [float, np.ndarray])
    42.24
    >>> validate_input('option1', str, ['option1', 'option2', 'option3'])
    'option1'
    >>> validate_input(-3.6, None, (None, 42))
    -3.6
    >>> validate_input(2.7, np.ndarray)
    array([2.7])
    >>> validate_input([1, 2, 3, 4.5], np.ndarray, [1, 2, 3, 4, 4.5, 5, 5.5])
    array([1. , 2. , 3. , 4.5])
    >>> validate_input((4.5, 6, 8), tuple, (0, 10))
    (4.5, 6, 8)
    >>> try:
    ...     validate_input([3, 2, 8], bool)
    ... except ValueError as err:
    ...     err.args[0]
    'The input value must be one of bool. It is list.'
    >>> try:
    ...     validate_input(72.1, float, (0, 13.2), 'asimov')
    ... except ValueError as err:
    ...     err.args[0]
    'The input value for asimov must be between 0 and 13.2. It is 72.1.'
    >>> try:
    ...     validate_input('option6', str, ['option1', 'option2', 'option3'])
    ... except ValueError as err:
    ...     err.args[0]
    'The input value must be one of option1, option2, option3. It is option6.'
    >>> try:
    ...     validate_input('option6', str, ['option1', 'option2', 'option3'])
    ... except ValueError as err:
    ...     err.args[0]
    'The input value must be one of option1, option2, option3. It is option6.'
    >>> try:
    ...     validate_input(2, int, [0, 1, 3, 9])
    ... except ValueError as err:
    ...     err.args[0]
    'The input value must be one of 0, 1, 3, 9. It is 2.'
    >>> try:
    ...     validate_input(np.array([5, -3, 7]), np.ndarray, (None, 3))
    ... except ValueError as err:
    ...     err.args[0]
    'The input value must be between -inf and 3. It is between -3 and 7.'
    >>> try:
    ...     validate_input(np.array([2, 9, 3]), np.ndarray, [0, 1, 3, 9])
    ... except ValueError as err:
    ...     err.args[0]
    'The input value must be one of 0, 1, 3, 9.'
    """
    if property_name is None:
        name = ""
    else:
        name = " for " + property_name
    if permissible_types is None:
        pass  # Disabled
    elif isinstance(permissible_types, List) and (
            isinstance(value, tuple(permissible_types))):
        pass  # Already correct
    elif not isinstance(permissible_types, List) and (
            isinstance(value, permissible_types)):
        pass  # Already correct
    elif (permissible_types in [bool, int, float, str]
          and isinstance(value, bool | int | float)):
        # Easy conversion
        value = permissible_types(value)
    elif permissible_types is np.ndarray and isinstance(
            value, bool | int | float):
        # Create array of length 1 with value
        value = np.atleast_1d(value)
    elif permissible_types is np.ndarray and isinstance(value, Iterable):
        # Convert List, Tuple, etc. to numpy array
        value = np.asarray(value)
    else:
        if not isinstance(permissible_types, List):
            permissible_types = [permissible_types]  # For unified error
        raise ValueError(
            f"The input value{name} must be one of "
            f"{','.join([p.__name__ for p in permissible_types])}. "
            f"It is {type(value).__name__}.")

    if permissible is None:
        return value
    elif isinstance(permissible, Tuple):
        if permissible[0] is None:
            permissible = (-np.inf, permissible[1])
        if permissible[1] is None:
            permissible = (permissible[0], np.inf)
    if isinstance(value, float | int):
        if isinstance(permissible, List):
            if value not in permissible:
                raise ValueError(
                    f"The input value{name} must be one of "
                    f"{', '.join([str(p) for p in permissible])}. "
                    f"It is {str(value)}.")
        elif isinstance(permissible, Tuple):
            if not (permissible[0] <= value <= permissible[1]):
                raise ValueError(
                    f"The input value{name} must be between "
                    f"{str(permissible[0])} and "
                    f"{str(permissible[1])}. It is {str(value)}.")
        return value
    elif isinstance(value, str):
        if value not in permissible:
            raise ValueError(
                f"The input value{name} must be one of "
                f"{', '.join([str(p) for p in permissible])}. "
                f"It is {value}.")
        else:
            return value
    elif isinstance(value, Iterable):
        if isinstance(permissible, List):
            if not all([v in permissible for v in value]):
                raise ValueError(
                    f"The input value{name} must be one of "
                    f"{', '.join([str(p) for p in permissible])}.")
        elif isinstance(permissible, Tuple):
            if isinstance(value, np.ndarray):  # Much faster for large arrays
                error = not (np.all(value >= permissible[0])
                             and np.all(value <= permissible[1]))
            else:
                error = not all([permissible[0] <= v <= permissible[1]
                                 for v in value])
            if error:
                raise ValueError(
                    f"The input value{name} must be between "
                    f"{str(permissible[0])} and {str(permissible[1])}. "
                    f"It is between {str(np.min(value))} and "
                    f"{str(np.max(value))}.")
        return value



"""
SysML Block
===========

This file implements a basic SysML Block.

References
----------
SysML Documentation:
    https://sysml.org/.res/docs/specs/OMGSysML-v1.4-15-06-03.pdf
"""

from __future__ import annotations

from typing import List, Any

import cetpy
from cetpy.Modules.Utilities.Labelling import name_2_abbreviation, \
    name_2_display
from cetpy.Modules.SysML import ValuePrinter
from cetpy.Modules.Solver import Solver


class Block:
    """SysML Block element."""

    __slots__ = ['_resetting', 'name', 'abbreviation', '_parent',
                 '_tolerance', '_logger',
                 'parts', 'ports', 'requirements', 'solvers',
                 '_get_init_parameters', '__init_kwargs__', '__dict__']

    __init_parameters__ = []
    __init_parts__ = []
    _reset_dict = {}
    _hard_reset_dict = {}
    _bool_parent_reset = True

    print = ValuePrinter()

    def __init__(self, name: str, abbreviation: str = None,
                 parent: Block = None, tolerance: float = None, **kwargs
                 ) -> None:
        # region Logging
        if cetpy.active_session is not None:
            self._logger = cetpy.active_session.logger
        else:
            self._logger = None
        # endregion

        # region Solver Flags
        self._resetting = False
        self._parent = None
        # endregion

        # region Attributes
        self.name = name
        if abbreviation is None:
            abbreviation = name_2_abbreviation(name)
        self.abbreviation = abbreviation
        self.parent = parent

        # region Config Loading
        session = cetpy.active_session

        def get_parent_name_list(block: Block) -> List[str]:
            """Return list of names of parent objects."""
            if block.parent is None:
                return [block.name]
            else:
                return get_parent_name_list(block.parent) + [block.name]
        parent_names = get_parent_name_list(self)

        def get_parent_class_list(block: Block) -> List[str]:
            """Return list of class names of parent objects."""
            if block.parent is None:
                return [block.__class__.__name__.lower()]
            else:
                return get_parent_name_list(block.parent) + [
                    block.__class__.__name__.lower()]
        class_names = get_parent_class_list(self)
        type_name = str(type(self))

        parameter_strings = [
            '.'.join(parent_names[i:]) for i in range(-len(parent_names), 0)
            ] + [
            '.'.join(class_names[i:]) for i in range(-len(class_names), 0)
            ] + [
            '.'.join(type_name[
                     type_name.find('Modules') + 8:-2].split('.')[:-2]).lower()
        ]

        config_keys = cetpy.active_session.config_manager.config_keys

        def parameter(key: str):
            """Return parameter setting for a given key with prioritisation.

            Source Prioritisation:
            1. kwargs
            2. session config
            3. default config

            Key Name Prioritisation:
            1. Chained block names from the highest parent to the block
            2. Block name
            3. Chained block class names from the highest parent to the block
            4. Block class name
            5. Key

            Example:
                Scenario: Setting the radius of the Moon Concordia of the
                Planet Mandalore:

                1. mandalore.concordia.radius
                2. concordia.radius
                3. Planet.Moon.radius
                4. Moon.radius
                5. radius

            Please note, for comparison to kwargs, all '.' separators are
            replaced by '_'.
            """
            key_strings = [ps + '.' + key for ps in parameter_strings] + [key]
            in_kwargs = [k.replace('.', '_') for k in key_strings
                         if k.replace('.', '_') in kwargs.keys()]
            in_config = [k for k in key_strings if k in config_keys]
            if len(in_kwargs) > 0:
                key_load = in_kwargs[0]
                source = 'kwargs'
                val = kwargs.get(key_load)
            elif len(in_config) > 0:
                key_load = in_config[0]
                source = 'config'
                val = session.parameter(key_load)
            else:
                return None
            if isinstance(val, str | float | int | None):
                self._logger.log(
                    15,
                    f"{key} loaded as {key_load} from {source}: {str(val)}")
            else:
                self._logger.log(
                    15, f"{key} loaded as {key_load} from {source}")
            return val
        self._get_init_parameters = parameter
        self.__init_kwargs__ = kwargs

        for key in self.__init_parameters__:
            self._resetting = True  # Avoid unnecessary resets.
            getattr(type(self), key).__set__(self, parameter(key))
            self._resetting = False
        # endregion
        # endregion

        # region References
        self.parts: List[Block] = []
        self.ports = []
        self.requirements = []
        self.solvers: List[Solver] = []

        # region Part Initialisation
        for key in self.__init_parts__:
            try:
                self.__setattr__(key, parameter(key + '_model'))
            except KeyError:
                self.__setattr__(key, None)
        # endregion
        # endregion

        # region Tolerance
        self._tolerance = 0
        if tolerance is None:
            tolerance = cetpy.active_session.parameter('tolerance')
        self.tolerance = tolerance
        # endregion

    # region Block References
    @property
    def parent(self) -> (Block, None):
        return self._parent

    @parent.setter
    def parent(self, val: Block) -> None:
        if val is None and self._parent is not None:
            try:
                self._parent.parts.remove(self)
            except ValueError:
                pass
        elif val is not None and self not in val.parts:
            val.parts += [self]
        self._parent = val

    def __deep_getattr__(self, name: str) -> Any:
        """Get value from block or its parts, solvers, and ports."""
        if '.' not in name:
            return self.__getattribute__(name)
        else:
            name_split = name.split('.')
            return self.__getattribute__(name_split[0]).__deep_getattr__(
                '.'.join(name_split[1:]))

    def __deep_setattr__(self, name: str, val: Any) -> None:
        """Set value on block or its parts, solvers, and ports."""
        if '.' not in name:
            return self.__setattr__(name, val)
        else:
            name_split = name.split('.')
            self.__getattribute__(name_split[0]).__deep_setattr__(
                '.'.join(name_split[1:]), val)
    # endregion

    # region Solver Functions
    def reset(self, parent_reset: bool = None) -> None:
        """Tell the instance and sub-blocks to resolve before the next
        value output."""
        if not self._resetting:
            self._resetting = True
            # reset all sub solvers while not calling the instances own reset
            for s in self.solvers:
                s.reset(parent_reset=False)

            # Reset all local attributes to the desired reset value
            for key, val in self._reset_dict.items():
                self.__setattr__(key, val)

            # Reset parent instance if desired:
            if parent_reset is None:
                parent_reset = self._bool_parent_reset
            if parent_reset and self.parent is not None:
                self.parent.reset()

            self._resetting = False

    def hard_reset(self, convergence_reset: bool = False) -> None:
        """Reset all blocks including solver flags and intermediate values.

        This should only be used to get the program unstuck as it undermines
        recursion stops and deletes any progress made.
        """
        self._resetting = False
        for key, val in self._hard_reset_dict.items():
            self.__setattr__(key, val)
        for solver in self.solvers:
            solver.hard_reset(convergence_reset)
        self.reset()

    @property
    def solved_self(self) -> bool:
        """Return bool if the block and its solvers are solved."""
        return True and (len(self.solvers) == 0
                         or all([s.solved for s in self.solvers]))

    @property
    def solved(self) -> bool:
        """Return bool if the block, its solvers, and its contained blocks
        are solved."""
        return self.solved_self and (len(self.parts) == 0
                                     or all([p.solved for p in self.parts]))

    def solve_self(self) -> None:
        """Solve the block and its solvers."""
        [s.solve() for s in self.solvers]

    def solve(self) -> None:
        """Solve the block, its solvers, and its parts."""
        self.solve_self()
        [p.solve() for p in self.parts]
    # endregion

    # region Input Properties
    @property
    def tolerance(self) -> float:
        """Block solver tolerances."""
        return self._tolerance

    @tolerance.setter
    def tolerance(self, val: float) -> None:
        if val < self._tolerance:
            self.reset()
        self._tolerance = val
        for p in [p for p in self.parts if p.tolerance > val]:
            p.tolerance = val
        for s in [s for s in self.solvers if s.tolerance > val]:
            s.tolerance = val
        for p in [p for p in self.ports if p.tolerance > val]:
            p.tolerance = val
    # endregion

    # region User Output
    @property
    def name_display(self) -> str:
        """Return a display formatted version of the block name."""
        return name_2_display(self.name)
    # endregion



"""
Generic Fluid Block
===================

This file specialises the FluidBlock to add generic properties common for
most fluid elements.

The model is intended for steady-state incompressible flow and is quite rough.
"""
from __future__ import annotations

import numpy as np

from cetpy.Modules.SysML import value_property, DeterminationTest
from cetpy.Modules.FluidBlock import FluidBlock


class GenericFluidBlock(FluidBlock):
    """Generic Fluid Block element."""

    dp_fixed = FluidBlock.dp_fixed
    dp_fixed.determination_test = DeterminationTest()

    __init_parameters__ = FluidBlock.__init_parameters__ + [
        'kv', 'cd',
    ]

    # region Transfer Functions
    def _dp_solve(self) -> float:
        """Pressure difference across the fluid element [Pa].

        This is the private calculate function.

        See Also
        --------
        FluidBlock.dp: public function with necessity check
        """
        if self.__class__.kv.fixed(self):
            return 1e5 * self.inlet.rho / 1000 / (
                    self.kv / (self.inlet.vdot * 3600)) ** 2
        elif self.__class__.cd.fixed(self) and self.area is not None:
            return (self.inlet.mdot / self.cda) ** 2 / (2 * self.inlet.rho)
        elif self.__class__.loss_factor.fixed(self):
            return self.loss_factor * self.inlet.rho / 2 * self.inlet.q ** 2
        else:
            return self.dp_fixed
    # endregion

    # region Input Properties
    @value_property(
        determination_test=FluidBlock.hydraulic_diameter.determination_test)
    def area(self) -> float:
        """Fluid element characteristic flow area [m^2]."""
        if self.__class__.hydraulic_diameter.fixed(self):
            return np.pi * (self.hydraulic_diameter / 2) ** 2
        elif self.__class__.cd.fixed(self):
            return self.cda / self.cd
        else:
            raise ValueError(
                "Neither hydraulic diameter, area, or discharge coefficient "
                "are set. An area cannot be calculated.")

    @value_property(determination_test=dp_fixed.determination_test)
    def kv(self) -> float:
        """Fluid element flow coefficient [m^3/h]."""
        return self.inlet.vdot * 3600 * np.sqrt(
            1e5 / self.dp * self.inlet.rho / 1000)

    @value_property(determination_test=dp_fixed.determination_test)
    def cd(self) -> float:
        """Fluid element discharge coefficient [-]."""
        return self.cda / self.area

    @value_property()
    def cda(self) -> float:
        """Fluid element effective flow area [m^2]."""
        if self.__class__.cd.fixed(self) and self.area is not None:
            return self.cd * self.area
        else:
            return self.inlet.mdot / np.sqrt(self.dp * self.inlet.rho * 2)

    @cda.setter
    def cda(self, val: float | None) -> None:
        if val is None:
            self.cd = None
        else:
            self.cd = val / self.area

    @value_property(determination_test=dp_fixed.determination_test)
    def loss_factor(self) -> float:
        """Fluid element loss coefficient [-]."""
        return 2 * self.dp / self.inlet.rho / self.inlet.q ** 2
    # endregion



"""
Class Skeleton for a generic Fluid
==================================

This file specifies the skeleton of a Fluid, that is the base parameters that
every Fluid implementation should implement.
"""

import numpy as np


class FluidSkeleton:
    """Elemental functions of all Fluid class implementations."""

    __slots__ = ['_name', '_function']

    def __init__(self, name: str, function: str = None):
        self.name = name
        self.function = function

    @property
    def name(self) -> str:
        """Fluid name (human-readable)."""
        return self._name

    @name.setter
    def name(self, val: str) -> None:
        self._name = val

    @property
    def function(self) -> str:
        """Fluid function in a system."""
        return self._function

    @function.setter
    def function(self, val: str | None) -> None:
        self._function = val

    def rho(self, t: float | np.ndarray, p: float | np.ndarray
            ) -> float | np.ndarray:
        """Return fluid density from temperature and pressure [kg/m^3]."""
        raise NotImplementedError

    def p(self, t: float | np.ndarray, rho: float | np.ndarray
          ) -> float | np.ndarray:
        """Return fluid pressure from temperature and density [Pa]."""
        raise NotImplementedError

    def h(self, t: float | np.ndarray, p: float | np.ndarray
          ) -> float | np.ndarray:
        """Return fluid mass specific enthalpy from temperature and
        pressure [J/kg]."""
        raise NotImplementedError

    def t_kappa(self, t1: float | np.ndarray, p1: float | np.ndarray,
                p2: float | np.ndarray) -> float | np.ndarray:
        """Return fluid isentropic expansion temperature from initial
         temperature and pressure and final pressure [K]."""
        raise NotImplementedError

    def t_boil(self, p: float | np.ndarray) -> float | np.ndarray:
        """Return fluid boiling temperature from pressure [K]."""
        raise NotImplementedError

    def p_boil(self, t: float | np.ndarray) -> float | np.ndarray:
        """Return fluid boiling pressure from temperature [Pa]."""
        raise NotImplementedError

    def t_crit(self) -> float:
        """Return critical temperature [K]."""
        raise NotImplementedError

    def p_crit(self) -> float:
        """Return critical pressure [Pa]."""
        raise NotImplementedError

    def phase(self, t: float | np.ndarray, p: float | np.ndarray
              ) -> str | np.ndarray:
        """Return fluid phase from temperature and pressure."""
        raise NotImplementedError

    def is_critical(self, t: float | np.ndarray, p: float | np.ndarray
                    ) -> float | np.ndarray:
        """Return bool true if fluid is supercritical."""
        return self.phase(t, p) in ['supercritical',
                                    'supercritical_liquid',
                                    'supercritical_gas']

    def is_gas(self, t: float | np.ndarray, p: float | np.ndarray
               ) -> float | np.ndarray:
        """Return bool true if fluid is gas or supercritical gas."""
        return self.phase(t, p) in ['gas', 'supercritical',
                                    'supercritical_gas']

    def t_rhoh(self, rho: float | np.ndarray, h: float | np.ndarray
               ) -> float | np.ndarray:
        """Return fluid temperature from density and enthalpy [K]."""
        raise NotImplementedError

    def t_ph(self, p: float | np.ndarray, h: float | np.ndarray
             ) -> float | np.ndarray:
        """Return fluid temperature from pressure and enthalpy [K]."""
        raise NotImplementedError

    def mu(self, t: float | np.ndarray, p: float | np.ndarray
           ) -> float | np.ndarray:
        """Return fluid dynamic viscosity from temperature and
        pressure [Pas]."""
        raise NotImplementedError

    def nu(self, t: float | np.ndarray, p: float | np.ndarray
           ) -> float | np.ndarray:
        """Return fluid kinematic viscosity from temperature and
        pressure [m^2/s]."""
        return self.mu(t, p) / self.rho(t, p)

    def c(self, t: float | np.ndarray, p: float | np.ndarray
          ) -> float | np.ndarray:
        """Return fluid mass specific heat capacity from temperature and
         pressure [J/(kgK)]."""
        raise NotImplementedError

    def a(self, t: float | np.ndarray, p: float | np.ndarray
          ) -> float | np.ndarray:
        """Return fluid speed of sound from temperature and pressure [m/s]."""
        raise NotImplementedError

    def pr(self, t: float | np.ndarray, p: float | np.ndarray
           ) -> float | np.ndarray:
        """Return fluid prandtl number from temperature and pressure [-]."""
        raise NotImplementedError

    def k(self, t: float | np.ndarray, p: float | np.ndarray
          ) -> float | np.ndarray:
        """Return fluid thermal conductivity from temperature and
        pressure [W/(mK)]."""
        raise NotImplementedError

    def kappa(self, t: float | np.ndarray, p: float | np.ndarray
              ) -> float | np.ndarray:
        """Return fluid isentropic expansion coefficient from temperature and
         pressure [-]."""
        raise NotImplementedError

    def m(self) -> float | np.ndarray:
        """Return fluid molecular mass [kg/mol]."""
        raise NotImplementedError



"""
SysML PartProperty
==================

This file implements a basic SysML Part Property. Moreover it handles the
setting and replacing of part properties, ensuring the references are
correctly broken and created and providing a more convenient interface to
selecting the desired model.

References
----------
SysML Documentation:
    https://sysml.org/.res/docs/specs/OMGSysML-v1.4-15-06-03.pdf
"""

from __future__ import annotations

from typing import Any, List, Tuple

from cetpy.Modules.Utilities.Labelling import name_2_display
import cetpy.Configuration


class PartPropertyDoc:
    """Simple Descriptor to pass a more detailed doc string to a part
    property."""

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        else:
            # noinspection PyProtectedMember
            return instance._super_class.__doc__

    def __set__(self, instance, value: str) -> None:
        if instance is not None and instance.fget is not None:
            instance.fget.__doc__ = value


class PartProperty:
    """SysML ValueProperty which adds units of measure and error
    calculation."""
    __slots__ = ['_name', '_name_instance', '_super_class', '_allowed_list']

    __doc__ = PartPropertyDoc()

    def __init__(self, super_class: type(object),
                 allowed_list: List[str | type(object) | None] = None) -> None:
        self._name = ''
        self._name_instance = ''

        self._super_class = super_class
        if allowed_list is None:
            allowed_list = [super_class]
        if any([isinstance(a, str) for a in allowed_list]):
            for i_a, a in allowed_list:
                if isinstance(a, str):
                    allowed_list[i_a] = cetpy.Configuration.get_module(a)
        self._allowed_list = allowed_list

    # region Getters and Setters
    def __set_name__(self, instance, name):
        self._name = name
        self._name_instance = '_' + name

    def str(self, instance) -> str:
        """Return formatted display name of the part."""
        return self.value(instance).name_display

    def value(self, instance) -> Any:
        """Return the part property value."""
        return self.__get__(instance)

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        name = self._name_instance
        if name in instance.__slots__:
            return instance.__getattribute__(name)
        else:
            return instance.__dict__.get(name, None)

    def __set__(self, instance, value) -> None:
        if value is None:
            if None in self._allowed_list:
                instance.__setattr__(self._name_instance, None)
                instance.reset()
                return
            value = self._super_class
        if isinstance(value, str):
            match_list = [
                cls for cls in self._allowed_list
                if value.lower().replace('_', '') == cls.__name__.lower()]
            if len(match_list) > 0:
                value = match_list[0]
            else:
                allowed_string = ', '.join([
                    cls.__name__.lower() for cls in self._allowed_list])
                raise ValueError(
                    f"The desired class could not be evaluated from the "
                    f"input {value}. Allowed classes are: {allowed_string}")
        if self._super_class in value.mro():
            initial_part = self.__get__(instance)
            part = value(name=self._name, parent=instance,
                         **instance.__init_kwargs__)
            instance.__setattr__(self._name_instance, part)
            if initial_part is not None:
                # noinspection PyUnresolvedReferences, PyDunderSlots
                initial_part.parent = None
                self._super_class.__set_state__(
                    part, self._super_class.__get_state__(initial_part))
            instance.reset()
        else:
            raise ValueError(f"The part property input is not an instance of "
                             f"the allowed super class {self._super_class}, "
                             f"instead it is {type(value)}.")
    # endregion

    # region Labelling
    @property
    def name_display(self) -> str:
        """Return the display formatted name of the value property."""
        return name_2_display(self._name)
    # endregion


class PartsProperty(PartProperty):
    """Extension of the cetpy SysML PartProperty class additionally
    incorporating support for multiplicity. Value returns are aggregate
    values of the individual parts."""
    __slots__ = ['_multiplicity']

    def __init__(self, super_class: type(object),
                 allowed_list: List[str | type(object)] = None,
                 multiplicity: int | List[int | None]
                 | Tuple[int | None] | str | None = None
                 ) -> None:
        super().__init__(super_class, allowed_list)
        self.multiplicity = multiplicity

    # region Getters and Setters
    def __set__(self, instance, value) -> None:
        if value is None:
            instance.__setattr__(self._name_instance, [])
            instance.reset()
            return
        if not isinstance(value, list):
            value = [value]
        for i_v, val in enumerate(value):
            if isinstance(val, str):
                match_list = [
                    cls for cls in self._allowed_list
                    if val.lower().replace('_', '') == cls.__name__.lower()]
                if len(match_list) > 0:
                    value[i_v] = match_list[0]
                else:
                    allowed_string = ', '.join([
                        cls.__name__.lower() for cls in self._allowed_list])
                    raise ValueError(
                        f"The desired class could not be evaluated from the "
                        f"input {val}. Allowed classes are: {allowed_string}")

        if not all([self._super_class not in val.mro() for val in value]):
            raise ValueError(f"The parts property input are not instances of "
                             f"the allowed super class {self._super_class}, "
                             f"instead they are "
                             f"{[type(val) for val in value]}.")

        initial_parts = self.__get__(instance)
        parts = []
        for val in value:
            parts += [val(name=self._name, parent=instance,
                          **instance.__init_kwargs__)]
        instance.__setattr__(self._name_instance, parts)
        for i_p, i_part in enumerate(initial_parts):
            # noinspection PyUnresolvedReferences, PyDunderSlots
            i_part.parent = None
            if len(parts) > i_part:
                self._super_class.__set_state__(
                    parts[i_p], self._super_class.__get_state__(i_part))
        instance.reset()
    # endregion

    # region Input Properties
    @property
    def multiplicity(self) -> int | List[int | None] | Tuple[int | None]:
        """Allowed number of instances of a given part property."""
        return self._multiplicity

    @multiplicity.setter
    def multiplicity(self, val: int | List[int | None]
                     | Tuple[int | None] | str | None) -> None:
        if isinstance(val, str):
            if '.' not in val and ',' not in val:
                if val == '*':
                    val = None
                else:
                    val = int(val)
            elif ',' in val:
                val = [int(v) if v != '*' else None for v in val.split(',')]
            elif '..' in val:
                val = tuple([
                    int(v) if v != '*' else None for v in val.split('..')])
            else:
                raise ValueError(
                    "Input string could not be parsed. allowed are 0, "
                    "positive integers, '*' for infinity, and the separators "
                    "',' for a list of possibilities, and '..' for lower and "
                    "upper bounds.")
        if val is None:
            self._multiplicity = 1
        elif isinstance(val, int):
            self._multiplicity = val
        elif isinstance(val, tuple) and len(val) == 2 and (
                isinstance(val[0], int | None)
                and isinstance(val[1], int | None)):
            self._multiplicity = val
        elif isinstance(val, list) and all([
                isinstance(val[i], int | None) for i in range(len(val))]):
            self._multiplicity = val
        else:
            raise ValueError("Invalid multiplicity input. Input must be "
                             "either None (== 1), an integer, list of "
                             "integers, or a string e.g. '1', '0,4', '0..*'.")
    # endregion



"""
CoolProp Implementation of a CET Fluid
======================================

This file specifies an implementation of a CET Fluid using the open-source
CoolProp Package.
"""

import numpy as np
from CoolProp.CoolProp import PropsSI, PhaseSI

from cetpy.Modules.Fluid import FluidSkeleton


class FluidCoolProp(FluidSkeleton):
    """Implements a CET Fluid using CoolProp."""

    __slots__ = ['_cpid']

    def __init__(self, name: str, function: str = None, cpid: str = None):
        self._cpid = cpid
        self._name = None  # for comparison with cpid.
        super().__init__(name, function)

    @property
    def name(self) -> str:
        """Fluid name (human-readable).

        Inherited function is extended to write the coolprop ID (cpid) if
        the cpid is currently not set or the same as the existing name."""
        return self._name

    @name.setter
    def name(self, val: str) -> None:
        if self._cpid is None or self._cpid == self._name:
            self._cpid = val
        self._name = val

    @property
    def cpid(self) -> str:
        """Fluid CoolProp ID."""
        return self._cpid

    @cpid.setter
    def cpid(self, val: str) -> None:
        self._cpid = val

    def rho(self, t: float | np.ndarray, p: float | np.ndarray
            ) -> float | np.ndarray:
        return PropsSI('D', 'T', t, 'P', p, self.cpid)

    def p(self, t: float | np.ndarray, rho: float | np.ndarray
          ) -> float | np.ndarray:
        return PropsSI('P', 'T', t, 'D', rho, self.cpid)

    def h(self, t: float | np.ndarray, p: float | np.ndarray
          ) -> float | np.ndarray:
        return PropsSI('H', 'T', t, 'P', p, self.cpid)

    def t_kappa(self, t1: float | np.ndarray, p1: float | np.ndarray,
                p2: float | np.ndarray) -> float | np.ndarray:
        h = self.h(t1, p1)
        return PropsSI('T', 'P', p2, 'H', h, self.cpid)

    def t_boil(self, p: float | np.ndarray) -> float | np.ndarray:
        return PropsSI('T', 'P', p, 'Q', 0, self.cpid)

    def p_boil(self, t: float | np.ndarray) -> float | np.ndarray:
        return PropsSI('P', 'T', t, 'Q', 0, self.cpid)

    def t_crit(self) -> float:
        return PropsSI('Tcrit', self.cpid)

    def p_crit(self) -> float:
        return PropsSI('Pcrit', self.cpid)

    def phase(self, t: float | np.ndarray, p: float | np.ndarray
              ) -> str | np.ndarray:
        return PhaseSI('T', t, 'P', p, self.cpid)

    def t_rhoh(self, rho: float | np.ndarray, h: float | np.ndarray
               ) -> float | np.ndarray:
        return PropsSI('T', 'D', rho, 'H', h, self.cpid)

    def t_ph(self, p: float | np.ndarray, h: float | np.ndarray
             ) -> float | np.ndarray:
        return PropsSI('T', 'P', p, 'H', h, self.cpid)

    def mu(self, t: float | np.ndarray, p: float | np.ndarray
           ) -> float | np.ndarray:
        return PropsSI('V', 'T', t, 'P', p, self.cpid)

    def c(self, t: float | np.ndarray, p: float | np.ndarray
          ) -> float | np.ndarray:
        return PropsSI('C', 'T', t, 'P', p, self.cpid)

    def a(self, t: float | np.ndarray, p: float | np.ndarray
          ) -> float | np.ndarray:
        return PropsSI('A', 'T', t, 'P', p, self.cpid)

    def pr(self, t: float | np.ndarray, p: float | np.ndarray
           ) -> float | np.ndarray:
        return PropsSI('PRANDTL', 'T', t, 'P', p, self.cpid)

    def k(self, t: float | np.ndarray, p: float | np.ndarray
          ) -> float | np.ndarray:
        return PropsSI('CONDUCTIVITY', 'T', t, 'P', p, self.cpid)

    def kappa(self, t: float | np.ndarray, p: float | np.ndarray
              ) -> float | np.ndarray:
        raise NotImplementedError

    def m(self) -> float | np.ndarray:
        raise NotImplementedError



"""
SysML ValueProperty
===================

This file implements a basic SysML Value Property.

References
----------
SysML Documentation:
    https://sysml.org/.res/docs/specs/OMGSysML-v1.4-15-06-03.pdf
"""

from __future__ import annotations

from typing import Any, Tuple, Callable, List, Iterable, Sized
import logging
import numpy as np

from cetpy.Modules.Utilities.Labelling import name_2_unit, name_2_axis_label, \
    scale_value, name_2_display
from cetpy.Modules.Utilities.InputValidation import validate_input


def value_property(equation: str = None,
                   determination_test: DeterminationTest = None,
                   necessity_test: float = 0.1,
                   permissible_list: None | List | Tuple = None,
                   permissible_types_list: None | type | List = None
                   ) -> Callable[[Callable], ValueProperty]:
    """Decorator Factory to create ValueProperties from getter functions."""

    def decorator(func: Callable) -> ValueProperty:
        """Decorator to create a ValueProperty with metadata from getter."""
        prop = ValueProperty(fget=func, equation=equation,
                             determination_test=determination_test,
                             necessity_test=necessity_test,
                             permissible_list=permissible_list,
                             permissible_types_list=permissible_types_list)
        return prop

    return decorator


class UnitFloat(float):
    """Extension fo the python float to add a unit representation."""

    __slots__ = ['unit']

    def __new__(cls, value: float, unit: str = None):
        return float.__new__(cls, value)

    def __init__(self, value: float, unit: str = None):
        super().__init__()
        self.unit = unit

    def __str__(self) -> str:
        value, prefix = scale_value(self * 1.0)
        return str(value) + ' ' + prefix + self.unit

    def __repr__(self) -> str:
        return self.__str__()


class DeterminationTest:
    """Class to manage over- and under-determination of model inputs."""

    __slots__ = ['_num', 'auto_fix', 'properties']

    def __init__(self, properties: List[str] = None,
                 num: int = 1, auto_fix: bool = True) -> None:
        self._num = num
        self.auto_fix = auto_fix
        if properties is None:
            properties: List[str] = []
        self.properties = properties

    @property
    def num(self) -> int:
        """Number of properties allowed to be fixed."""
        return self._num

    @num.setter
    def num(self, val: int) -> None:
        self._num = int(val)

    def test(self, instance, new: str = None) -> None:
        """Test the determination of the instance. If auto fix is on,
        the test attempts to automatically fix it."""
        n_actual = sum([getattr(type(instance), n).fixed(instance)
                        for n in self.properties])
        n_target = self._num

        direction = ''
        amendment = ''
        if n_actual == n_target:
            return
        elif n_actual > n_target:
            direction = 'over-'
            if self.auto_fix and self._num == 1:
                [instance.__setattr__(n, None) for n in self.properties
                 if n != new]
                # noinspection PyUnresolvedReferences
                logging.info(f"Autocorrected {instance.name_display} "
                             f"over-determination. {new} is the new input.")
                return
            elif self.auto_fix:
                amendment = ' Autofix failed.'

        elif n_actual < n_target:
            direction = 'under-'

        # noinspection PyUnresolvedReferences
        logging.warning(
            f"{instance.name_display} is {direction}determined. Of "
            f"{', '.join(self.properties)} {n_target} must be set. "
            f"Currently {n_actual} are set.{amendment}")


class ValuePropertyDoc:
    """Simple Descriptor to pass a more detailed doc string."""

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        elif instance.fget is not None:
            return instance.fget.__doc__
        else:
            return instance.name_display + f"[{instance.unit}]."

    def __set__(self, instance, value: str) -> None:
        if instance is not None and instance.fget is not None:
            instance.fget.__doc__ = value


class ValueProperty:
    """SysML ValueProperty which adds units of measure and error
    calculation."""
    __slots__ = ['_name', '_name_instance', '_name_instance_reset',
                 '_determination_test', '_necessity_test', 'equation',
                 '_unit', '_axis_label', 'fget', 'fset', 'fdel',
                 '_permissible_list', '_permissible_types_list']

    __doc__ = ValuePropertyDoc()

    def __init__(self,
                 unit: str = None, axis_label: str = None,
                 fget: Callable = None, fset: Callable = None,
                 fdel: Callable = None, equation: str = None,
                 determination_test: DeterminationTest = None,
                 necessity_test: float = 0.1,
                 permissible_list: None | List | Tuple = None,
                 permissible_types_list: None | type | List = None) -> None:
        self._determination_test = None
        self._necessity_test = None
        self._permissible_list = None
        self._permissible_types_list = None
        self._name = ''
        self._name_instance = ''
        self._name_instance_reset = ''
        self.equation = equation
        self.determination_test = determination_test
        self.necessity_test = necessity_test
        self.permissible_list = permissible_list
        self.permissible_types_list = permissible_types_list
        if (unit is None and fget is not None
                and fget.__doc__ is not None and '].' in fget.__doc__):
            doc = fget.__doc__
            unit = doc[doc.find('[') + 1: doc.find('].')]
        self._unit = unit
        if (axis_label is None and fget is not None
                and fget.__doc__ is not None
                and '..cetpy:axis_label:' in fget.__doc__):
            doc = fget.__doc__
            axis_label = doc[
                doc.find('..cetpy:axis_label:') + 17: doc.find('\n')].strip()
        self._axis_label = axis_label

        self.fget = fget
        self.fset = fset
        self.fdel = fdel

    # region Decorators
    def getter(self, fget: Callable) -> ValueProperty:
        prop = self
        prop.fget = fget
        return prop

    # ToDo: Figure out type hinting for read access
    def setter(self, fset: Callable) -> ValueProperty:
        prop = self
        prop.fset = fset
        return prop

    def deleter(self, fdel: Callable) -> ValueProperty:
        prop = self
        prop.fdel = fdel
        return prop
    # endregion

    # region Getters and Setters
    def __set_name__(self, cls, name):
        self._name = name
        self._name_instance = '_' + name
        self._name_instance_reset = self._name_instance + '_reset'
        if self._unit is None:
            self._unit = name_2_unit(name)
        if self._axis_label is None:
            self._axis_label = name_2_axis_label(name)
        # Rerun determination test setter when the name is known
        self.determination_test = self._determination_test

    def name(self) -> str:
        """Value Property Name"""
        return self._name

    def str(self, instance) -> str:
        """Return formatted string of value."""
        value = self.value(instance)
        if isinstance(value, str):
            return value
        elif isinstance(value, (int, float)):
            value, prefix = scale_value(self.value(instance))
            unit = self.unit
            if any([str(i) in unit for i in range(10)]):
                unit = prefix + '(' + unit + ')'
            else:
                unit = prefix + unit
            return str(value) + ' ' + unit
        else:
            return str(value)

    def value(self, instance) -> Any:
        """Return the value property value with significant figure rounding."""
        value = self.value_raw(instance)
        if isinstance(value, float):
            return float('{:.{p}g}'.format(value, p=5))
        else:
            return value

    def value_raw(self, instance) -> Any:
        """Return the value property value without further modification."""
        return self.__get__(instance)

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        name = self._name_instance

        try:
            value = instance.__getattribute__(name)
        except AttributeError:
            value = None

        if value is not None or self.fget is None:
            return value
        else:
            return self.fget(instance)

    def __set__(self, instance, value) -> None:
        try:
            val_initial = instance.__getattribute__(self._name_instance_reset)
        except AttributeError:
            val_initial = None
        value = validate_input(value, self.permissible_types_list,
                               self.permissible_list, self._name)
        if self.fset is None:
            instance.__setattr__(self._name_instance, value)
        else:
            self.fset(instance, value)

        if self._determination_test is not None:
            self._determination_test.test(instance, self._name)

        if self.__reset_necessary__(instance, value, val_initial):
            instance.__setattr__(self._name_instance_reset, value)
            instance.reset()

    def __reset_necessary__(self, instance, value, val_initial=None) -> bool:
        """Return bool if a reset should be performed on an instance if the
        initial value is changed to value.
        """
        necessity_test = self.necessity_test
        reset = True
        if necessity_test >= 0:
            same = (value == val_initial)
            if isinstance(same, Iterable):
                same = all(same)
            if same:
                reset = False
            elif type(value) == type(val_initial):
                if isinstance(value, int | float):
                    reset = not np.isclose(
                        value, val_initial,
                        rtol=instance.tolerance * necessity_test, atol=0)
                if isinstance(value, Iterable | Sized) and (
                        len(value) == len(val_initial)):
                    reset = not all(np.isclose(
                        value, val_initial,
                        rtol=instance.tolerance * necessity_test, atol=0))
        return reset

    def __delete__(self, instance):
        if self.fdel is not None:
            self.fdel(instance)
        else:
            delattr(instance, self._name_instance)

    def fixed(self, instance) -> bool:
        """Return bool if value property is a fixed parameter for the given
        instance."""
        name = self._name_instance
        if instance is None:
            return False
        elif name in instance.__slots__:
            return instance.__getattribute__(name) is not None
        else:
            return instance.__dict__.get(name, None) is not None

    @property
    def determination_test(self) -> (DeterminationTest, None):
        """Over- and Under-Determination Test that applies to this value
        property."""
        return self._determination_test

    @determination_test.setter
    def determination_test(self, val: (DeterminationTest, None)) -> None:
        name = self._name
        if name != '':  # init call, rerun this function in __set_name__
            # Remove self from past test
            if self._determination_test is not None:
                try:
                    self._determination_test.properties.remove(name)
                except ValueError:  # Not in list.
                    pass
            # Add self to new test
            if val is not None and name not in val.properties:
                val.properties += [name]
        self._determination_test = val

    @property
    def necessity_test(self) -> float:
        """Reset necessity test multiplier. If this value property of an
        instance is changed by a larger relative factor than this multiplier
        times the instance tolerance, a reset is called. This multiplier is
        intended to counteract the error sensitivity of an input value.

        The default is 0.1, which is moderately conservative for a large number
        of models. In effect this means, that if a block is set to a
        tolerance of 1e-3, a 1e-4 tolerance is applied to the inputs.

        Evaluating the error sensitivity and calibrating this value can
        significantly speed up models.

        This setter does not automatically call a reset. The user has to
        manually reset all relevant models after changing this value.
        """
        return self._necessity_test

    @necessity_test.setter
    def necessity_test(self, val: float) -> None:
        if not isinstance(val, float | int) and val >= 0:
            raise ValueError("The necessity test multiplier must be a "
                             "positive float. Set to 0 to disable.")
        self._necessity_test = val

    @property
    def permissible_list(self) -> None | List | Tuple:
        """List of permissible values for the value property. Set to None to
        disable. Set to a list of approved values, particularly useful for
        string switches. Or set to a tuple of length two with integer or
        float lower and upper limits. None represents positive or negative
        infinity.
        """
        return self._permissible_list

    @permissible_list.setter
    def permissible_list(self, val: None | List | Tuple) -> None:
        if not (val is None or isinstance(val, List)
                or (isinstance(val, Tuple) and len(val) == 2
                    and all([isinstance(v, None | float | int) for v in val]))
                ):
            raise ValueError("The permissible list must be None or a "
                             "List or a Tuple of length 2 with float / None.")
        if isinstance(val, Tuple):
            if val[0] is None:
                val = (-np.inf, val[1])
            if val[1] is None:
                val = (val[0], np.inf)
            if val[1] < val[0]:
                raise ValueError("The allowed minimum limit is above the "
                                 "upper limit. The requirement cannot be "
                                 "fulfilled.")
        self._permissible_list = val

    @property
    def permissible_types_list(self) -> None | type | List[type]:
        """List of permissible types for input values. New inputs are tested
        against this list and an error is raised if it cannot be satisfied.
        Minor conversions are attempted automatically. These include float |
        int to bool and vice-versa or list to np array.
        """
        return self._permissible_types_list

    @permissible_types_list.setter
    def permissible_types_list(self, val: None | type | List[type]) -> None:
        if not isinstance(val, None | type | List):
            raise ValueError("The permissible types list must be None, "
                             "a type or a List of allowed types.")
        self._permissible_types_list = val
    # endregion

    # region Labelling
    @property
    def name_display(self) -> str:
        """Return the display formatted name of the value property."""
        return name_2_display(self._name)

    @property
    def unit(self) -> str:
        """Unit of measure."""
        return self._unit

    @unit.setter
    def unit(self, val: str) -> None:
        self._unit = val

    @property
    def axis_label(self) -> str:
        """Axis label for visualisation."""
        return self._axis_label

    @axis_label.setter
    def axis_label(self, val: str) -> None:
        self._axis_label = val
    # endregion


class ValuePrinter:
    """Formatter for block Value Properties to display the property in the
    truncated value form with unit of measure."""
    __slots__ = ['_instance']

    def __get__(self, instance, owner=None):
        self._instance = instance
        return self

    def __getattribute__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            vp = getattr(type(self._instance), name)
            if isinstance(vp, ValueProperty):
                return vp.str(self._instance)
            else:
                return str(vp.fget(self._instance))


class AggregateValueProperty(ValueProperty):
    """An extension of the SysML ValueProperty class to return the sum of
    the value property on the current block and all its contained parts."""

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        value = super().__get__(instance, owner)

        if value is None:  # Make compatible for addition with parts' values
            value = 0.

        for p in instance.parts:
            try:  # Try except should be faster than dir lookup.
                part_value = p.__getattribute__(self._name)
                if part_value is not None:
                    value = value + part_value  # add non-mutable
            except AttributeError:
                pass

        return value

    def __get_self__(self, instance, owner=None):
        return super().__get__(instance, owner)



"""
Case Runner
===========

This file defines a CET case runner designed to generate and execute
thousands of design or operating variations.
"""
from __future__ import annotations

import logging
from typing import List
from time import perf_counter
import numpy as np
import pandas as pd
from traceback import format_tb

import cetpy.Configuration
import cetpy.Modules.SysML
import cetpy.Modules.Solver


class CaseSolver(cetpy.Modules.Solver.Solver):

    def _solve_function(self) -> None:
        runner: CaseRunner = self.parent
        output_df = runner.output_df
        instances = runner.module_instances

        for case in output_df.loc[~output_df.solved, :].itertuples():
            i = int(getattr(case, 'Index'))
            logging.info(f'Executing case {i}.')
            print(f'Executing case {i}.')
            t1 = perf_counter()
            if runner.catch_errors:
                # noinspection PyBroadException
                try:
                    runner.__initialise_case__(case)
                    runner.__evaluate_case__(case)
                except Exception as err:
                    output_df.loc[i, 'solved'] = True
                    output_df.loc[i, 'errored'] = True
                    output_df.loc[i, 'error_class'] = err.__class__.__name__
                    output_df.loc[i, 'error_message'] = err.args[0]
                    output_df.loc[i, 'error_location'] = format_tb(
                        err.__traceback__)[-1]
                    output_df.loc[i, 'code_time'] = perf_counter() - t1
                    continue
            else:
                runner.__initialise_case__(case)
                runner.__evaluate_case__(case)
            output_df.loc[i, 'code_time'] = perf_counter() - t1

            if runner.save_instances:
                # noinspection PyProtectedMember
                instances[i] = runner._instance


class CaseRunner(cetpy.Modules.SysML.Block):
    """Congruent Engineering Toolbox Case Runner."""

    module = cetpy.Modules.SysML.ValueProperty(
        permissible_types_list=[type(cetpy.Modules.SysML.Block), str])
    input_df = cetpy.Modules.SysML.ValueProperty(
        permissible_types_list=pd.DataFrame)
    save_instances = cetpy.Modules.SysML.ValueProperty(
        permissible_types_list=bool)
    catch_errors = cetpy.Modules.SysML.ValueProperty(
        permissible_types_list=bool)
    additional_module_kwargs = cetpy.Modules.SysML.ValueProperty(
        permissible_types_list=[dict, type(None)])
    output_properties = cetpy.Modules.SysML.ValueProperty(
        permissible_types_list=[list, type(None)])

    __init_parameters__ = cetpy.Modules.SysML.Block.__init_parameters__ + [
        'module', 'input_df', 'save_instances', 'catch_errors',
        'additional_module_kwargs', 'output_properties'
    ]

    _reset_dict = cetpy.Modules.SysML.Block._reset_dict
    _reset_dict.update({'_output_df': None, '_module_instances': None,
                        '_instance': None})

    def __init__(self, module: cetpy.Modules.SysML.Block | str = None,
                 input_df: pd.DataFrame = None,
                 save_instances: bool = False,
                 catch_errors: bool = True,
                 additional_module_kwargs: dict = None,
                 output_properties: List[str] = None, **kwargs):
        super().__init__(kwargs.pop('name', 'case_runer'),
                         module=module, input_df=input_df,
                         save_instances=save_instances,
                         catch_errors=catch_errors,
                         additional_module_kwargs=additional_module_kwargs,
                         output_properties=output_properties,
                         **kwargs)
        self._output_df = None
        self._module_instances = None
        self._instance = None
        self.case_solver = CaseSolver(parent=self)

    @property
    def input_keys(self) -> List[str]:
        """List of keys in input_df."""
        return self.input_df.keys()

    @module.setter
    def module(self, val: cetpy.Modules.SysML.Block | str) -> None:
        if isinstance(val, str):
            self._module = cetpy.Configuration.get_module(val)
        else:
            self._module = val

    @property
    def output_df(self) -> pd.DataFrame:
        """Return completed run of cases with input followed by output
        values."""
        if self._output_df is None:
            # Initialise output dataframe from input dataframe and add
            # columns for solver progress, any occurring error diagnostic,
            # and performance timing.
            self._output_df = self.input_df.copy()
            cols = ['solved', 'errored', 'error_class', 'error_message',
                    'error_location']
            self._output_df.loc[:, cols] = False
            self._output_df.loc[:, 'code_time'] = np.nan
        self.case_solver.solve()
        return self._output_df

    @property
    def module_instances(self) -> List[cetpy.Modules.SysML.Block]:
        """Return list of solved module instances if save_instances is
        enabled."""
        if self._module_instances is None:
            self._module_instances = [None] * self.input_df.shape[0]
        self.case_solver.solve()
        return self._module_instances

    def __initialise_case__(self, case) -> None:
        """Initialise the instance for a single specific case."""
        kwargs = self.additional_module_kwargs.copy()
        if kwargs is None:
            kwargs = {}

        for col in [c for c in case._fields if c in self.input_keys]:
            kwargs.update({col: getattr(case, col)})

        # noinspection PyPropertyAccess
        self._instance = self.module(**kwargs)

    def __evaluate_case__(self, case) -> None:
        """Evaluate and write the output for a single_specific_case."""
        output_properties = self.output_properties
        df = self.output_df
        i = getattr(case, 'Index')
        instance = self._instance
        if output_properties is None:
            instance.solve()
            for vp in [type(instance).__getattr__(p)
                       for p in instance.__dir__() if isinstance(
                    type(instance).__getattr__(p),
                    cetpy.Modules.SysML.ValueProperty)]:
                df.loc[i, vp.name] = vp.__get__(instance)
        else:
            for col in output_properties:
                df.loc[i, col] = instance.__deep_getattr__(col)



"""
Class Skeleton for a generic Material
=====================================

This file specifies the skeleton of a generic Material, that is the base
parameters that every Material implementation should implement.
"""

from typing import List
import numpy as np


class MaterialSkeleton:
    """Elemental functions of all Material class implementations."""

    __slots__ = ['_name']

    def __init__(self, name: str):
        self.name = name

    @property
    def name(self) -> str:
        """Fluid name (human-readable)."""
        return self._name

    @name.setter
    def name(self, val: str) -> None:
        self._name = val

    def rho(self, t: float | np.ndarray | None = None) -> float | np.ndarray:
        """Return material density from temperature [kg/m^3].

        If no temperature is specified. Room Temperature 298.15 K (25 deg C)
        is assumed.
        """
        raise NotImplementedError

    def ultimate_tensile_strength(self, t: float | np.ndarray | None = None
                                  ) -> float | np.ndarray:
        """Return material ultimate tensile strength from temperature [kg/m^3].

        If no temperature is specified. Room Temperature 298.15 K (25 deg C)
        is assumed.
        """
        raise NotImplementedError

    def yield_strength(self, t: float | np.ndarray | None = None
                       ) -> float | np.ndarray:
        """Return material yield strength from temperature [kg/m^3].

        If no temperature is specified. Room Temperature 298.15 K (25 deg C)
        is assumed.
        """
        raise NotImplementedError

    def thermal_conductivity(self, t: float | np.ndarray | None = None) -> \
            float | np.ndarray:
        """Return material thermal conductivity from temperature [kg/m^3].

        If no temperature is specified. Room Temperature 298.15 K (25 deg C)
        is assumed.
        """
        raise NotImplementedError

    def thermal_expansion(self, t: float | np.ndarray | None = None
                          ) -> float | np.ndarray:
        """Return material thermal expansion coefficient from temperature [-].

        If no temperature is specified. Room Temperature 298.15 K (25 deg C)
        is assumed.
        """
        raise NotImplementedError

    def is_compatible(self, substance: str | List[str]) -> bool | List[bool]:
        """Return bool if fluid is compatible with a given substance."""
        raise NotImplementedError



"""
SysML Continuous Port
=====================

This file implements a SysML Port, an interface between multiple SysML
Blocks for Flow Items with relevant functions for continuous flow, such as
fluid flow.
"""

from __future__ import annotations

from typing import List

import cetpy.Modules.SysML as SML


class ContinuousPort(SML.Port):
    """SysML Port element adding functions for continuous flow systems,
    such as fluid flow."""

    __slots__ = ['_upstream_list', '_upstream_port_list', '_downstream_list',
                 '_downstream_port_list']

    def __init__(self,
                 upstream: SML.Block | None = None,
                 downstream: SML.Block | None = None,
                 name: str = None, flow_item: object | None = None,
                 upstream_dict_name: str = None,
                 downstream_dict_name: str = None,
                 tolerance: float = None):
        self._upstream_list = None
        self._upstream_port_list = None
        self._downstream_list = None
        self._downstream_port_list = None
        super().__init__(upstream=upstream,
                         downstream=downstream,
                         name=name,
                         flow_item=flow_item,
                         upstream_dict_name=upstream_dict_name,
                         downstream_dict_name=downstream_dict_name)

    # region System References
    @SML.Port.upstream.setter
    def upstream(self, val: SML.Block | None) -> None:
        SML.Port.upstream.fset(self, val)
        self.__generate_upstream_list__()
        for port in self.downstream_port_list:
            port.__generate_upstream_list__()

    @SML.Port.downstream.setter
    def downstream(self, val: SML.Block | None) -> None:
        SML.Port.downstream.fset(self, val)
        self.__generate_downstream_list__()
        for port in self.upstream_port_list:
            port.__generate_downstream_list__()

    def _update_reference(self, reference: str,
                          val: SML.Block | None):
        val_initial = self.__getattribute__(reference)
        SML.Port._update_reference(self, reference, val)

        if val is val_initial:
            return
        # Update all flow system lists
        self.__generate_upstream_list__()
        self.__generate_downstream_list__()

        for port in self.flow_system_port_list:
            port.__generate_downstream_list__()
            port.__generate_upstream_list__()

        # Update flow item from possible opposite port. This must be done
        # after all new references are established.

        # Step 1: Identify opposite port:
        if reference == '_upstream':
            opposite_name = '_downstream'
        else:
            opposite_name = '_upstream'
        try:
            opposite = val.__getattribute__(self.__getattribute__(
                opposite_name + '_dict_name'))
        except AttributeError:
            opposite = None

        if opposite is None:
            return  # Do nothing
        elif self._flow_item is None and opposite.flow_item is not None:
            # Step 2: Attempt to update own and previously attached elements'
            # flow items using new flow item.
            self.flow_item = opposite.flow_item
        elif self._flow_item is not None and opposite.flow_item is None:
            # Step 3: Attempt to update opposite and its previously attached
            # elements' flow items using own flow item.
            opposite.flow_item = self._flow_item

        # create transfer functions, only if opposite exists (not an end point)
        for fp in self.__flow_properties__:
            fp.__create_block_delta_attribute__(val)

    @property
    def upstream_list(self) -> List[SML.Block]:
        """Return list of upstream blocks with this flow from upstream to
        downstream."""
        if self._upstream_list is None:
            self.__generate_upstream_list__()
        return self._upstream_list

    @property
    def downstream_list(self) -> List[SML.Block]:
        """Return list of downstream blocks with this flow from upstream to
        downstream."""
        if self._downstream_list is None:
            self.__generate_downstream_list__()
        return self._downstream_list

    @property
    def flow_system_list(self) -> List[SML.Block]:
        """Return list of blocks with this flow from upstream to downstream."""
        return self.upstream_list + self.downstream_list

    @property
    def upstream_port_list(self) -> List[ContinuousPort]:
        """Return list of upstream ports from upstream to downstream."""
        if self._upstream_port_list is None:
            self.__generate_upstream_list__()
        return self._upstream_port_list

    @property
    def downstream_port_list(self) -> List[ContinuousPort]:
        """Return list of downstream ports from upstream to downstream."""
        if self._downstream_port_list is None:
            self.__generate_downstream_list__()
        return self._downstream_port_list

    @property
    def flow_system_port_list(self) -> List[ContinuousPort]:
        """Return list of ports from upstream to downstream."""
        return self.upstream_port_list + [self] + self.downstream_port_list

    def __generate_upstream_list__(self) -> None:
        """Update the upstream port list."""
        if self._upstream is None:
            self._upstream_list = []
            self._upstream_port_list = []
        else:
            try:
                next_port = self._upstream.__getattribute__(
                    self._downstream_dict_name)
                self._upstream_list = \
                    next_port.upstream_list + [self._upstream]
                self._upstream_port_list = \
                    next_port.upstream_port_list + [next_port]
            except AttributeError:
                self._upstream_list = [self._upstream]
                self._upstream_port_list = []

    def __generate_downstream_list__(self) -> None:
        """Update the upstream port list."""
        if self._downstream is None:
            self._downstream_list = []
            self._downstream_port_list = []
        else:
            try:
                next_port = self._downstream.__getattribute__(
                    self._upstream_dict_name)
                self._downstream_list = \
                    [self._downstream] + next_port.downstream_list
                self._downstream_port_list = \
                    [next_port] + next_port.downstream_port_list
            except AttributeError:
                self._downstream_list = [self._downstream]
                self._downstream_port_list = []

    @property
    def is_endpoint(self) -> bool:
        port_list = self.flow_system_port_list
        return self is port_list[0] or self is port_list[-1]

    @property
    def is_upstream_endpoint(self) -> bool:
        return self is self.flow_system_port_list[0]

    @property
    def is_downstream_endpoint(self) -> bool:
        return self is self.flow_system_port_list[-1]
    # endregion

    # region Input Properties
    @SML.Port.flow_item.setter
    def flow_item(self, val: object | None) -> None:
        SML.Port.flow_item.fset(self, val)
        # Additionally see if adjacent ports should be updated. This must be
        # done after the own flow item is updated.
        for port in self.upstream_port_list[::-1]:
            if port.flow_item is not None:  # Does not update any ports that
                # have a flow item set and breaks loop.
                break
            # Reset on all ports already called in super() setter method.
            port.__setattr__('_flow_item', self._flow_item)
        for port in self.downstream_port_list:
            if port.flow_item is not None:
                break
            port.__setattr__('_flow_item', self._flow_item)
    # endregion



"""
SysML Port
==========

This file implements a basic SysML Port, an interface between multiple SysML
Blocks for Flow Items.
"""

from __future__ import annotations
from typing import Any

import cetpy.Modules.SysML


class Port:
    """SysML Port element."""
    __slots__ = ['_upstream', '_downstream', '_upstream_dict_name',
                 '_downstream_dict_name', 'name', 'parent', '_flow_item',
                 '_resetting', '_tolerance', '__dict__']

    __flow_properties__ = []

    def __init__(self,
                 upstream: cetpy.Modules.SysML.Block | None = None,
                 downstream: cetpy.Modules.SysML.Block | None = None,
                 name: str = None, flow_item: object | None = None,
                 upstream_dict_name: str = None,
                 downstream_dict_name: str = None,
                 tolerance: float = None):
        self._resetting = False
        self._flow_item = None
        self._upstream = None
        self._downstream = None
        self.name = name
        self._upstream = upstream
        self._downstream = downstream
        self._upstream_dict_name = upstream_dict_name
        self._downstream_dict_name = downstream_dict_name
        self._flow_item = flow_item
        self._tolerance = 0
        if tolerance is None:
            tolerance = 0
        self.tolerance = tolerance

        for fp in self.__flow_properties__:
            fp.__init_private_attributes__(self)

    # region System References
    @property
    def upstream(self) -> cetpy.Modules.SysML.Block | None:
        """Port upstream element."""
        return self._upstream

    @upstream.setter
    def upstream(self, val: cetpy.Modules.SysML.Block | None):
        self._update_reference('_upstream', val)

    @property
    def downstream(self) -> cetpy.Modules.SysML.Block | None:
        """Port downstream element."""
        return self._downstream

    @downstream.setter
    def downstream(self, val: cetpy.Modules.SysML.Block | None):
        self._update_reference('_downstream', val)

    def _update_reference(self, reference: str,
                          val: cetpy.Modules.SysML.Block | None):
        """Update a specified system reference."""
        val_initial = self.__getattribute__(reference)
        dict_name = self.__getattribute__(reference + '_dict_name')
        if val is val_initial:
            return
        elif val_initial is not None:
            val_initial.ports.remove(self)
            port_copy = self.__copy_unlinked__()
            val_initial.ports += [port_copy]
            val_initial.__setattr__(dict_name, port_copy)
            if reference == '_upstream':
                opposite_name = '_downstream'
            else:
                opposite_name = '_upstream'
            port_copy.__setattr__(opposite_name, val_initial)
            val_initial.reset()
        if val is not None:
            val.ports += [self]
            val.__setattr__(dict_name, self)
            val.reset()
        self.__setattr__(reference, val)

    @property
    def is_endpoint(self) -> bool:
        """Return bool if the port is either a starting or ending node."""
        return True

    @property
    def is_upstream_endpoint(self) -> bool:
        """Return bool if the port is a starting node."""
        return True

    @property
    def is_downstream_endpoint(self) -> bool:
        """Return bool if the port is an ending node."""
        return True

    def __copy_unlinked__(self) -> Port:
        """Return a copy of the port without up- or downstream references."""
        port = type(self)(name=self.name,
                          flow_item=self.flow_item,
                          upstream_dict_name=self._upstream_dict_name,
                          downstream_dict_name=self._downstream_dict_name,
                          tolerance=self.tolerance)
        for fp in self.__flow_properties__:
            # noinspection PyProtectedMember
            for prop in [fp._name_instance, fp._name_instance_input,
                         fp._name_instance_reset, fp._name_instance_direction]:
                port.__setattr__(prop, self.__getattribute__(prop))
        return port

    def __deep_getattr__(self, name: str) -> Any:
        """Get value from block or its parts, solvers, and ports."""
        if '.' not in name:
            return self.__getattribute__(name)
        else:
            name_split = name.split('.')
            return self.__getattribute__(name_split[0]).__deep_getattr__(
                '.'.join(name_split[1:]))

    def __deep_setattr__(self, name: str, val: Any) -> None:
        """Set value on block or its parts, solvers, and ports."""
        if '.' not in name:
            return self.__setattr__(name, val)
        else:
            name_split = name.split('.')
            self.__getattribute__(name_split[0]).__deep_setattr__(
                '.'.join(name_split[1:]), val)
    # endregion

    # region Resetting
    def reset(self) -> None:
        """Reset port, up- and downstream elements."""
        if not self._resetting:
            self._resetting = True
            self.reset_upstream()
            self.reset_downstream()
            # reset self is called by both up- and down-stream resets
            self._resetting = False

    def reset_upstream(self) -> None:
        """Reset port and upstream elements."""
        self.reset_self()
        if self._upstream is not None:
            self._upstream.reset()

    def reset_downstream(self) -> None:
        """Reset port and downstream elements."""
        self.reset_self()
        if self._downstream is not None:
            self._downstream.reset()

    def reset_self(self) -> None:
        """Reset port stored intermediate values."""
        pass
    # endregion

    # region Input Properties
    @property
    def flow_item(self) -> object | None:
        """Port flow item. This object describes the behaviour of the flow."""
        return self._flow_item

    @flow_item.setter
    def flow_item(self, val: object | None) -> None:
        initial_val = self._flow_item
        if initial_val is val:
            return
        else:
            self._flow_item = val
            self.reset()

    @property
    def tolerance(self) -> float:
        """Port solver tolerances."""
        return self._tolerance

    @tolerance.setter
    def tolerance(self, val: float) -> None:
        if val < self._tolerance:
            self.reset()
        self._tolerance = val
    # endregion



"""
Configuration Manager
=====================

This file defines a Configuration Manager for CET. The manager handles the
interface to the .toml config files, layering of config files,
and traceability.
"""

from os import listdir
from os.path import join, isdir, isfile, abspath, dirname
from typing import List
import pandas as pd
import numpy as np
import pickle
import tomli

import cetpy
import cetpy.Configuration


def get_absolute_path(file_path: str, additional_locations: List[str] = None
                      ) -> str:
    """Return absolute path of a file within the CET config structure.

    A valid absolute path is prioritised above all else. Second the
    additional locations in order of occurrence, and lastly the CET config
    locations from cetpy.Configurations in order of occurrence.
    """
    if isfile(abspath(file_path)):
        return abspath(file_path)
    else:

        for location in (additional_locations +
                         cetpy.Configuration.config_locations):
            abs_path = join(location, file_path)
            if isfile(abs_path):
                return abs_path


def load_config(file_path: str) -> dict:
    """Return dictionary fitting to a TOML file at the specified location."""
    with open(file_path, "rb") as f:
        return tomli.load(f)


def interpret_config(config: dict, additional_locations: List[str] = None
                     ) -> dict:
    """Parse a config dictionary to load numpy, pickle, or csv files."""
    if additional_locations is None:
        additional_locations = []
    out = config.copy()
    for key in out.keys():
        if isinstance(out[key], dict) and 'file' in out[key].keys():
            match out[key]['file'][-3:]:
                case 'csv':
                    out[key] = pd.read_csv(get_absolute_path(
                        out[key]['file'], additional_locations))
                case '.pkl':
                    with open(out[key]['file'], "rb") as f:
                        out[key] = pickle.load(f)
                case '.txt':
                    out[key] = np.loadtxt(out[key]['file'])
                case _:
                    pass
        elif isinstance(out[key], dict):
            out[key] = interpret_config(out[key], additional_locations)
        elif isinstance(out[key], str) and out[key] in ['none', 'None']:
            out[key] = None
    return out


def get_parameter_dict() -> dict:
    """Return parameter dict incorporating all possible parameters
    across all model elements."""
    out = load_config(
        join(dirname(__file__), 'default_config_parameters.toml'))

    def load_parameters_from_dir(file_path: str) -> dict:
        target_dict = {}
        for item in listdir(file_path):
            if isdir(join(file_path, item)):
                sub_dict = load_parameters_from_dir(
                    join(file_path, item))
                if len(sub_dict) > 0:
                    target_dict[item.lower()] = sub_dict
            if isfile(join(file_path, item)) and item[-5:] == '.toml':
                target_dict.update(load_config(join(file_path, item)))
        return target_dict

    for loc in cetpy.Configuration.module_locations:
        out.update(load_parameters_from_dir(loc))
    return out


def get_default_dict(parameter_dict: dict) -> dict:
    """Return a default parameter dict based on a parameter dict."""
    out = {}
    for key, value in parameter_dict.items():
        if isinstance(value, dict) and 'default' in value.keys():
            out[key] = value['default']
        elif isinstance(value, dict):
            out[key] = get_default_dict(value)
    return out


def get_config_keys(config: dict) -> List[str]:
    """Return list of all value keys in the config, split by a dot,
    as would be the case in a config file."""
    keys = list(config.keys())
    for key, val in config.items():
        if isinstance(val, dict):
            keys += [key + '.' + k for k in get_config_keys(val)]
    return keys


def get_parameter(config: dict, key: str | List[str]):
    """Return the parameter value from a config dictionary using either the
    chained name (table headers split by a '.') or a list of keys."""
    if isinstance(key, str):
        key = key.split('.')
    val = config[key[0]]
    if len(key) > 1:
        return get_parameter(val, key[1:])
    else:
        return val


def format_config_header(header: str) -> List[str]:
    """Format the Title header of a config file."""
    lines = ["=" * 80 + "\n"]
    lines += [header.center(80) + "\n"]
    lines += ["=" * 80 + '\n\n']
    return lines


def format_config_table_header(table_level: List[str]) -> List[str]:
    """Format the table header of a config file"""
    return ["[" + ".".join(table_level) + "]"]


def format_value_to_toml(value) -> str:
    """Convert a config value to a print string for a .toml config file."""
    if isinstance(value, float | int | bool):
        return str(value)
    elif isinstance(value, str):
        return "'" + value + "'"
    elif isinstance(value, list):
        return "[" + ",".join([format_value_to_toml(v) for v in value]) + "]"
    elif isinstance(value, dict):
        return "{" + ",".join([f"{k} = {v}" for k, v in value.items()]) + "}"
    else:
        raise ValueError(
            f"Value type could not be converted to toml: {type(value)}")


def format_config_key(key: str, value, parameter_entry: dict) -> List[str]:
    """Format the table header of a config file"""
    value = format_value_to_toml(value)
    lines = [key + ' = ' + value]
    match parameter_entry.get('type', None), parameter_entry.get('unit', None):
        case None, None:
            pass
        case _, None:
            lines += ["(" + parameter_entry['type'] + ")"]
        case None, _:
            lines += ["[" + parameter_entry['unit'] + "]"]
        case _, _:
            lines += ["(" + parameter_entry['type'] + ") ["
                      + parameter_entry['unit'] + "]"]
        case _:
            pass

    lines += [parameter_entry['description']]
    return lines


def format_toml_body_text(default_dict: dict, parameter_dict: dict,
                          table_master_level: list = None) -> List[str]:
    """Format the body text of a TOML config file."""
    if table_master_level is None:
        table_master_level = []
    lines = []

    keys_values = [key for key, value in default_dict.items()
                   if not isinstance(value, dict)]
    keys_sub_keys = [key for key in default_dict.keys() if key not in
                     keys_values]
    for key in keys_values:
        lines += format_config_key(
            key, default_dict[key], parameter_dict[key])
    lines += ["\n\n"]

    for key in keys_sub_keys:
        table_sub_level = table_master_level + [key]
        lines += format_config_table_header(table_sub_level)
        parameter_sub_dict = parameter_dict[key]
        default_sub_dict = default_dict[key]
        lines += format_toml_body_text(default_sub_dict, parameter_sub_dict,
                                       table_sub_level)
    return lines


class ConfigurationManager:
    """CET Config Manger for loading, layering, and tracing configs."""

    __slots__ = ['_parameter_dict', '_config', '_configs', '_config_keys',
                 '_directory', '_session_config_name']

    def __init__(self, directory: str | None = None,
                 session_config: str | None = None) -> None:
        self._parameter_dict = {}
        self._config = {}
        self._configs = {}
        self._config_keys = []
        self.session_config_name = session_config
        self.directory = directory

    def reset(self) -> None:
        """Clear all loaded configs."""
        self._parameter_dict = {}
        self._config = {}
        self._configs = {}

    @property
    def config(self) -> dict:
        """Session config dictionary defining system and program parameters."""
        return self._config

    @property
    def directory(self) -> str | None:
        """Active directory for the config manager."""
        return self._directory

    @directory.setter
    def directory(self, val: str) -> None:
        self._directory = val
        self.load()

    @property
    def session_config_name(self) -> str:
        """Name of the session config file. If none is specified,
        the default 'session_config.toml' is used."""
        if self._session_config_name is None:
            return "session_config.toml"
        else:
            return self._session_config_name

    @session_config_name.setter
    def session_config_name(self, val: str | None) -> None:
        self._session_config_name = val

    @property
    def parameter_dict(self) -> dict:
        """Return dictionary of parameter properties, e.g. unit, data type,
        and description."""
        if len(self._parameter_dict) == 0:
            self._parameter_dict = get_parameter_dict()
        return self._parameter_dict

    @property
    def default_dict_raw(self) -> dict:
        """Return uninterpreted version of the default dictionary."""
        return get_default_dict(self.parameter_dict)

    @property
    def config_keys(self) -> List[str]:
        """Return list of all value keys in the config, split by a dot,
        as would be the case in a config file."""
        return self._config_keys

    def parameter(self, key: str | List[str]):
        """Return config parameter based on chained key string split by '.'
        or list of keys."""
        return get_parameter(self._config, key)

    def __get_abs_path__(self, file_path: str) -> str:
        """Return absolute path of a file within the CET config structure."""
        return get_absolute_path(file_path, [self.directory])

    def __get_config__(self, file_path: str) -> dict:
        """Return dictionary fitting to a TOML file at the specified location
        while finding a relevant config in the CET config structure."""
        return load_config(self.__get_abs_path__(file_path))

    def __load_config_amendment__(self, config: dict,
                                  config_dict: dict = None,
                                  allow_user_directory: bool = True
                                  ) -> dict:
        """Overload a config with its stated amendments."""
        if 'config_amendment' in config.keys():
            config_in = config.copy()
            config_amendment = config_in['config_amendment']
            if not isinstance(config_amendment, list):
                config_amendment = [config_amendment]
            for cam in config_amendment:
                if allow_user_directory:
                    config_add = self.__get_config__(cam)
                else:
                    config_add = load_config(get_absolute_path(cam))
                # Add config to list first, to preserve initial state
                # and prioritisation.
                if config_dict is not None:
                    config_dict[cam] = config_add
                config_add = self.__load_config_amendment__(
                    config_add, config_dict, allow_user_directory)
                config_in.update(config_add)
        return config

    def load(self) -> None:
        """Load the configuration at the specified directory."""
        # Always start from zero for a clean load.
        # region Default Config
        config = self.default_dict_raw
        self._configs = {'default': config.copy()}
        self._config = self.__load_config_amendment__(
            config, self._configs, False)
        # endregion

        # region Session Config
        if self.directory is not None:
            config = self.__get_config__(self.session_config_name)
            self._configs['session'] = config.copy()
            self._config.update(self.__load_config_amendment__(
                config, self._configs, True))

        self._config = interpret_config(self._config, [self.directory])
        # endregion

        self._config_keys = get_config_keys(self.config)

    def write_session_config_template(self, file_path: str = None) -> None:
        """Write template for the session config. The session config
        documents the available options with all installed modules, units,
        acceptable inputs, and parameter descriptions."""
        directory = self.directory
        if directory is None and file_path is None:
            raise ValueError("The config manager needs a defined directory "
                             "to write the session config template.")
        if file_path is None and isfile(
                join(directory, 'session_config.toml')):
            raise FileExistsError("Session config already exists. Delete "
                                  "this file first before writing again.")
        if file_path is None:
            file_path = join(directory, 'session_config.toml'),

        # region Write Text
        lines = format_config_header(
            "Congruent Engineering Toolbox Session Config")
        lines += format_toml_body_text(
            self.default_dict_raw, self.parameter_dict)
        # endregion

        with open(file_path, 'w') as f:
            [f.write("#" + line + '\n') for line in lines]



"""
Labelling Functions
===================

Provides a number of functions for predicting abbreviations, units, labels,
name formatting based on block or value property names.
"""

import re
import numpy as np


def name_2_abbreviation(name: str) -> str:
    """Return a sensible abbreviation for a component name."""
    name_split = name.split('_')
    if len(name_split) == 1:
        abbreviation = name_split[0][:3].upper()
    else:
        abbreviation = "".join([n[0].upper() for n in name_split[:3]])
    suffix = re.findall(r'^[A-Za-z0-9]*?([0-9]+)$', name)
    if len(suffix) > 0:
        abbreviation += suffix[0]
    return abbreviation


def name_2_display(name: str) -> str:
    """Return a display formatted version of a block name."""
    return ' '.join([w.title() if w.islower() else w for w in name.split('_')])


# standard units, keys are the trigger and values the unit string
standard_units = {
    'eta': '-',
    'error': '-',
    'ratio': '-',
    'dp': 'Pa',
    'prandtl': '-',
    'dt': 'K',
    'dh': 'J/kg',
    'angle': '-',
    'gamma': '-',
    'kappa': '-',
    'mach': '-',
    'radius': 'm',
    'diameter': 'm',
    'length': 'm',
    'area': 'm^2',
    'volume': 'm^3',
    'thickness': 'm',
    'height': 'm',
    'width': 'm',
    'pressure': 'Pa',
    'stress': 'Pa',
    'temperature': 'K',
    'density': 'kg/m^3',
    'rho': 'kg/m^3',
    'dynamic viscosity': 'Pa*s',
    'kinematic viscosity': 'm^2/s',
    'specific heat capacity': 'J/kgK',
    'thermal conductivity': 'W/mK',
    'enthalpy': 'J/kg',
    'entropy': 'J/kg',
    'heat flux': 'W/m^2',
    'heat transfer coefficient': 'W/m^2K',
    'velocity': 'm/s',
    'mass flow': 'kg/s',
    'mass': 'kg',
    'volume flow': 'm^3/s',
    'gas constant': 'J/kgK',
    'force': 'N',
    'torque': 'Nm',
    'power': 'W',
    'energy': 'J',
    'voltage': 'V',
    'charge': 'C',
    'magnetic flux': 'Wb',
    'capacitance': 'F',
    'resistance': 'ohm',
    'magnetic induction': 'T',
    'frequency': 'Hz',
    'rpm': '1/min',
    'index': '-',
    'kv': 'm^3/h',
    'mw': 'kg/kmol',
    'mu': 'Pa*s',
    'nu': 'm^2/s',
    'cp': 'J/kgK',
    'cv': 'J/kgK',
    'x': 'm',
    'y': 'm',
    'z': 'm'
}


def name_2_unit(name: str) -> str:
    """Return unit from property name."""
    keys = [k for k in standard_units.keys()
            if k in name.lower().replace('_', ' ')]
    if len(keys) > 0:
        return standard_units[keys[0]]
    else:
        return '-'


def unit_2_latex(unit: str) -> str:
    """Return Latex formatted unit."""
    # Ensure numbers are raised to the power
    unit_copy = ''
    for char in unit:
        if char in [str(n) for n in range(10)]:
            if (len(unit_copy) == 0
                    or unit_copy[-1] in [
                        str(n) for n in range(10)] + ['^']):
                unit_copy += char
            else:
                unit_copy += '^' + char
        else:
            unit_copy += char
    unit = unit_copy
    # Detect division
    if '/' in unit:
        unit = unit.split('/')
        unit = r'\frac{' + unit[0] + '}{' + unit[1] + '}'

    # Enable math mode
    return '$' + unit + '$'


# standard axis labels, keys are the trigger and values the display string
standard_labels = ['ratio', 'error', 'radius', 'diameter', 'length', 'area',
                   'volume', 'thickness', 'height', 'width', 'stress',
                   'pressure', 'temperature', 'density',
                   'specific heat capacity', 'thermal conductivity',
                   'enthalpy', 'entropy', 'heat flux',
                   'heat transfer coefficient', 'velocity', 'gas constant',
                   'prandtl', 'force', 'torque', 'power', 'energy', 'voltage',
                   'charge', 'magnetic flux', 'capacitance', 'resistance',
                   'magnetic induction', 'frequency', 'index']
standard_labels = dict(zip(standard_labels,
                           [label.title() for label in standard_labels]))
standard_labels.update({
    'rpm': 'RPM',
    'dp': 'Pressure Delta',
    'dt': 'Temperature Delta',
    'dh': 'Enthalpy Delta',
    'eta': r'$\eta$',
    'angle': r'$\theta$',
    'gamma': r'$\gamma$',
    'kappa': r'$\kappa$',
    'rho': r'$\rho$',
    'mu': r'$\mu$',
    'nu': r'$\nu$',
    'cp': r'$c_P$',
    'cv': r'$c_V$',
    'mw': r'$M_W$',
    'dynamic viscosity': r'$\mu$',
    'kinematic viscosity': r'$\nu$',
    'mach': 'Mach Number',
    'x': 'x-Coordinate',
    'y': 'y-Coordinate',
    'z': 'z-Coordinate'
})


def name_2_axis_label(name: str) -> str:
    """Return axis label from property name."""
    if 'eta' in name:  # Prioritise efficiency
        return standard_labels['eta']
    elif 'ratio' in name:  # Prioritise ratio
        return standard_labels['ratio']
    else:
        keys = [k for k in standard_labels.keys() if k in name]
        if len(keys) > 0:
            return standard_labels[keys[0]]
        else:
            return name.title().replace('_', ' ')


unit_abbreviations = {
    'm': 'meters',
    'kg': 'kilogram',
    'J': 'Joules',
    'K': 'Kelvin',
    'Pa': 'Pascal',
    'mol': 'Mole',
    's': 'Seconds',
    'A': 'Ampere',
    'Hz': 'Hertz',
    'C': 'Coulomb',
    'cd': 'Candela',
    'V': 'Volt',
    'W': 'Watt',
    'N': 'Newton',
    'rad': 'Radian',
    'F': 'Farad',
    'ohm': 'Ohm',
    'Wb': 'Weber',
    'T': 'Tesla'
}
si_base_units = ['s', 'm', 'kg', 'A', 'K', 'mol', 'cd']
si_units = ['m', 'kg', 'J', 'K', 'Pa', 'mol', 's', 'A', 'cd']
si_units_signs = si_units + ['/', '^', '(', ')'] + [str(n) for n in range(10)]

unit_base_equivalences = {
    'J': 'kgm^2s^-3',
    'Pa': 'kgm^-1s^-2',
    'Hz': 's^-1',
    'C': 'sA',
    'V': 'kgm^2s^-3A^-1',
    'W': 'kgm^2s^-3',
    'N': 'kgms^-2',
    'rad': '',
    'F': 'kg^-1m^-2s^4A^2',
    'ohm': 'kgm^2s^4A^2',
    'Wb': 'kgm^2s^-2A^-1',
    'T': 'kgs^-2A^-1'
}


def unit_2_baseSI(unit: str) -> str:
    """Convert a unit to base SI units."""
    unit_out = unit
    for key, value in unit_base_equivalences:
        unit_out.replace(key, value)
    return unit_out


standard_unit_magnitude_abbreviations = [
    'E', 'P', 'T', 'G', 'M', 'k', 'h', 'da', '', 'd', 'c', 'm', 'mu', 'n',
    'p', 'f'
]
standard_unit_magnitude_long_form = [
    'Exo', 'Peta', 'Tera', 'Giga', 'Mega', 'kilo', 'hecta', 'deka', '', 'deci',
    'centi', 'mili', 'micro', 'nano', 'pico', 'femto'
]
standard_unit_magnitudes = np.array([
    1e18, 1e15, 1e12, 1e9, 1e6, 1e3, 1e2, 1e1, 1, 1e-1, 1e-2, 1e-3, 1e-6, 1e-9,
    1e-12, 1e-15
])


def scale_value(value: float) -> (float, str):
    """Apply base 3 scaling prefixes to a value and return new value/prefix."""
    index_min = np.where((value > standard_unit_magnitudes))[0][0]
    if index_min in [6, 7, 9, 10]:  # restrict to base 3
        index_min = 8
    value = value / standard_unit_magnitudes[index_min]
    return value, standard_unit_magnitude_abbreviations[index_min]



"""
Compressible Fluid Block
========================

This file specialises the FluidBlock to add a more specific temperature
delta function to calculate temperature changes in gases with expansion or
compression.

The model is intended for steady-state flow and is quite rough.
"""

from cetpy.Modules.FluidBlock import FluidBlock


class CompressibleFluidBlock(FluidBlock):
    """Compressible Fluid Block element."""

    # region Transfer Functions
    def _dt_solve(self) -> float:
        """Temperature difference across the fluid element [K].

        This is the private calculate function.

        The function is expanded from the base FluidBlock to calculate the
        temperature rise from compression or expansion of a gas.

        See Also
        --------
        FluidBlock.dt: public function with necessity check
        """
        if self.dt_fixed == 0:
            return self.dt_fixed
        else:
            # Calculate temperature change with expansion / compression.
            t_out = self.inlet.flow_item.t_kappa(self.inlet.t, self.inlet.p,
                                                 self.outlet.p)
            return t_out - self.inlet.t
    # endregion



"""
Fluid Port
==========

This file implements an extension of the Continuous Port which specialises
for a fluid port. This adds an area property, and properties for pressure,
temperature, and mass-flow which together define the state of the fluid at
the port.
"""

from __future__ import annotations
import numpy as np
from copy import copy

import cetpy
import cetpy.Modules.SysML as SML
from cetpy.Modules.SysML import value_property
from cetpy.Modules.Fluid import FluidSkeleton


class FluidPort(SML.ContinuousPort):
    """SysML Port element adding functions for continuous flow systems
    containing specific flow properties and output value properties
    tailored for fluid flow."""

    p = SML.FlowProperty()
    t = SML.FlowProperty()
    mdot = SML.FlowProperty()
    area_mode = SML.ValueProperty(
        '', permissible_list=['fixed', 'upstream', 'downstream'],
        permissible_types_list=str)

    def __init__(self,
                 upstream: SML.Block | None = None,
                 downstream: SML.Block | None = None,
                 name: str = None,
                 flow_item: FluidSkeleton | None = None,
                 area_mode: str = None,
                 area: float = None,
                 pressure: float = None,
                 temperature: float = None,
                 mass_flow: float = None) -> None:
        super().__init__(upstream=upstream,
                         downstream=downstream,
                         name=name,
                         flow_item=flow_item,
                         upstream_dict_name='_outlet',
                         downstream_dict_name='_inlet')
        self._area_mode = 'fixed'
        self._area = None
        if area_mode is None and area is None:
            area_mode = 'upstream'
        elif area_mode is None:
            area_mode = 'fixed'
        self.area = area
        self.area_mode = area_mode  # After area to preserve user mode choice
        if pressure is not None:
            self.p = pressure
        else:
            self.__class__.p.__set_converging_value__(self, 10e5)
        if temperature is not None:
            self.t = temperature
        else:
            self.__class__.t.__set_converging_value__(self, 298.15)
        if mass_flow is not None:
            self.mdot = mass_flow
        else:
            self.__class__.mdot.__set_converging_value__(self, 0)

    # region System References
    def _update_reference(self, reference: str,
                          val: SML.Block | None):
        val_initial = self.__getattribute__(reference)
        SML.ContinuousPort._update_reference(self, reference, val)

        if val is val_initial:
            return
        if val_initial is not None:
            # noinspection PyUnresolvedReferences
            fluid_solver_copy = copy(val.fluid_solver)
            fluid_blocks_initial = val_initial.inlet_no_solve.flow_system_list
            fluid_solver_copy.parent = val_initial
            fluid_solver_copy.parents = fluid_blocks_initial
            for fb in fluid_blocks_initial:
                fb.fluid_solver = fluid_solver_copy
                # noinspection PyUnresolvedReferences
                fb.solvers.remove(val.fluid_solver)
                fb.solvers += [fluid_solver_copy]

        if val is not None:
            # noinspection PyUnresolvedReferences
            fluid_blocks = val.inlet_no_solve.flow_system_list
            # noinspection PyUnresolvedReferences
            fluid_solver = val.fluid_solver
            fluid_solver.parent = val
            fluid_solver.parents = fluid_blocks
            for fb in fluid_blocks:
                fb.fluid_solver = fluid_solver
                if fluid_solver not in fb.solvers:
                    fb.solvers += [fluid_solver]

    def __copy_unlinked__(self) -> FluidPort:
        # noinspection PyPropertyAccess
        port = FluidPort(name=self.name,
                         flow_item=self.flow_item,
                         area_mode=self.area_mode,
                         area=self.area)
        for fp in self.__flow_properties__:
            # noinspection PyProtectedMember
            for prop in [fp._name_instance, fp._name_instance_input,
                         fp._name_instance_reset, fp._name_instance_direction]:
                port.__setattr__(prop, self.__getattribute__(prop))
        return port
    # endregion

    # region Input Properties
    @SML.ContinuousPort.flow_item.setter
    def flow_item(self, val: FluidSkeleton | str | None) -> None:
        if isinstance(val, str):
            val = cetpy.Modules.Fluid.Fluid(val)
        SML.ContinuousPort.flow_item.fset(self, val)

    @value_property()
    def area(self) -> float:
        """Port flow area [m^2]."""
        if self.area_mode == 'fixed':
            return self._area
        elif self.area_mode == 'upstream' and self._upstream is not None:
            return self._upstream.__getattribute__('area')
        elif self.area_mode == 'upstream' and self._downstream is not None:
            return self._downstream.__getattribute__('area')
        elif self.area_mode == 'downstream' and self._downstream is not None:
            return self._downstream.__getattribute__('area')

    @area.setter
    def area(self, val: float) -> None:
        self._area = val
        if val is not None:
            self.area_mode = 'fixed'

    # noinspection PyPropertyAccess
    @value_property()
    def hydraulic_diameter(self) -> float:
        """Port hydraulic diameter [m]."""
        return np.sqrt(self.area / np.pi) * 2

    @hydraulic_diameter.setter
    def hydraulic_diameter(self, val: float) -> None:
        self.area = np.pi * (val / 2) ** 2
    # endregion

    # region Fluid Output Properties
    # noinspection PyPropertyAccess
    @value_property()
    def q(self) -> float:
        """Fluid flow velocity [m/s]."""
        return self.vdot / self.area

    @value_property()
    def p_s(self) -> float:
        """Fluid static pressure [Pa]."""
        return self.p - 0.5 * self.q ** 2 * self.rho

    @value_property()
    def t_s(self) -> float:
        """Fluid static temperature [K]."""
        return self.t - 0.5 * self.q ** 2 / self.c

    @value_property()
    def rho(self) -> float:
        """Fluid density [kg/m^3].

        This value is based on total temperature and pressure, due to model
        accuracy and complexity. You can verify the error with Port.rho_s.

        See Also
        --------
        FluidPort.rho_s: Density based on static temperature and pressure.
        """
        return self.flow_item.rho(self.t, self.p)

    @value_property()
    def rho_s(self) -> float:
        """Fluid density based on static properties [kg/m^3]."""
        return self.flow_item.rho(self.t_s, self.p_s)

    @value_property()
    def vdot(self) -> float:
        """Port volume flow [m^3/s]."""
        return self.mdot / self.rho

    @vdot.setter
    def vdot(self, val: float) -> None:
        self.mdot = val * self.rho

    # noinspection PyPropertyAccess
    @value_property()
    def re(self) -> float:
        """Fluid Reynold's Number [-]."""
        return self.q * self.hydraulic_diameter / self.nu

    @value_property()
    def h(self) -> float:
        """Fluid total enthalpy [J/kg]."""
        return self.flow_item.h(self.t, self.p)

    @value_property()
    def h_s(self) -> float:
        """Fluid static enthalpy [J/kg]."""
        return self.flow_item.h(self.t_s, self.p_s)

    @value_property()
    def phase(self) -> str:
        """Fluid phase name."""
        return self.flow_item.phase(self.t, self.p)

    @value_property()
    def mu(self) -> float:
        """Fluid dynamic viscosity [Pas]."""
        return self.flow_item.mu(self.t, self.p)

    @value_property()
    def nu(self) -> float:
        """Fluid kinematic viscosity [m^2/s]."""
        return self.flow_item.nu(self.t, self.p)

    @value_property()
    def c(self) -> float:
        """Fluid specific heat capacity [J/(kgK)]."""
        return self.flow_item.c(self.t, self.p)

    @value_property()
    def a(self) -> float:
        """Fluid speed of sound [m/s]."""
        return self.flow_item.a(self.t, self.p)

    @value_property()
    def pr(self) -> float:
        """Fluid Prandtl number [-]."""
        return self.flow_item.pr(self.t, self.p)

    @value_property()
    def k(self) -> float:
        """Fluid thermal conductivity [W/(mK)]."""
        return self.flow_item.k(self.t, self.p)

    @value_property()
    def kappa(self) -> float:
        """Fluid isentropic expansion coefficient [-]."""
        return self.flow_item.kappa(self.t, self.p)

    @value_property()
    def m(self) -> float:
        """Fluid molecular mass [J/kg]."""
        return self.flow_item.m()
    # endregion



"""
Implementation of a Material with given parameters
==================================================

This file specifies an implementation of a Material with given parameters.
"""

from typing import List
import numpy as np

from cetpy.Modules.Material._MaterialSkeleton import MaterialSkeleton


class MaterialGiven(MaterialSkeleton):
    """CET Material with given properties."""

    __slots__ = ['_rho', '_ultimate_tensile_strength', '_yield_strength',
                 '_yield_strength', '_thermal_conductivity',
                 '_thermal_expansion', '_compatibility']

    def __init__(self, name: str, rho: float = None,
                 ultimate_tensile_strength: float = None,
                 yield_strength: float = None,
                 thermal_conductivity: float = None,
                 thermal_expansion: float = None,
                 compatibility: List[str] = None):
        super().__init__(name)
        self._rho = rho
        self._ultimate_tensile_strength = ultimate_tensile_strength
        self._yield_strength = yield_strength
        self._thermal_conductivity = thermal_conductivity
        self._thermal_expansion = thermal_expansion
        self._compatibility = compatibility

    def rho(self, t: float | np.ndarray | None = None) -> float | np.ndarray:
        return self._rho

    def ultimate_tensile_strength(self, t: float | np.ndarray | None = None
                                  ) -> float | np.ndarray:
        return self._ultimate_tensile_strength

    def yield_strength(self, t: float | np.ndarray | None = None
                       ) -> float | np.ndarray:
        return self._yield_strength

    def thermal_conductivity(self, t: float | np.ndarray | None = None
                             ) -> float | np.ndarray:
        return self._thermal_conductivity

    def thermal_expansion(self, t: float | np.ndarray | None = None
                          ) -> float | np.ndarray:
        return self._thermal_expansion

    def is_compatible(self, substance: str | List[str]) -> bool | List[bool]:
        if isinstance(substance, str):
            return substance in self._compatibility
        else:
            return [s in self._compatibility for s in substance]



"""
Generic Decentralised Solver
============================

This file specifies the base Solver class which defines the base structure
of the decentralised solver architecture.
"""

from typing import List, Any

from cetpy.Modules.SysML import ValuePrinter


class Solver:
    """Decentralised Solver of the Congruent Engineering Toolbox."""

    __slots__ = ['_recalculate', '_calculating', '_hold', '_resetting',
                 '_name', 'parent', 'convergence_keys', '_tolerance']

    print = ValuePrinter()

    def __init__(self, parent, tolerance: float = None):
        self._recalculate = True
        self._calculating = False
        self._resetting = False
        self._hold = 0
        self.convergence_keys: List[str] = []
        self.parent = parent
        if parent is not None and self not in parent.solvers:
            parent.solvers += [self]
        self._tolerance = 0
        if tolerance is None and self.parent is not None:
            tolerance = self.parent.tolerance
        self.tolerance = tolerance

    # region Interface Functions
    def __set_name__(self, instance, name):
        self._name = name
        self.parent = instance
        if self not in instance.solvers:
            instance.solvers += [self]

    def reset(self, parent_reset: bool = True) -> None:
        """Tell the solver to resolve before the next value output."""
        if not self._resetting:
            self._resetting = True
            self._recalculate = True
            # Reset parent instance if desired
            if parent_reset:
                self.parent.reset()
            self._resetting = False

    def hard_reset(self, convergence_reset: bool = False) -> None:
        """Reset the in progress solver flags and call a normal reset."""
        self._resetting = False
        self._calculating = False
        self.reset()
        if convergence_reset:
            for key in self.convergence_keys:
                self.__setattr__(key, None)

    def __deep_getattr__(self, name: str) -> Any:
        """Get value from block or its parts, solvers, and ports."""
        if '.' not in name:
            return self.__getattribute__(name)
        else:
            name_split = name.split('.')
            return self.__getattribute__(name_split[0]).__deep_getattr__(
                '.'.join(name_split[1:]))

    def __deep_setattr__(self, name: str, val: Any) -> None:
        """Set value on block or its parts, solvers, and ports."""
        if '.' not in name:
            return self.__setattr__(name, val)
        else:
            name_split = name.split('.')
            self.__getattribute__(name_split[0]).__deep_setattr__(
                '.'.join(name_split[1:]), val)
    # endregion

    # region Solver Flags
    @property
    def solved(self) -> bool:
        """Return bool if the solver is solved."""
        return not self._recalculate

    @property
    def calculating(self) -> bool:
        """Return bool if the solver is currently running."""
        return self._calculating

    @property
    def solved_calculating(self) -> bool:
        """Return bool if the solver is solved or currently running."""
        return self.solved or self.calculating

    @property
    def hold(self) -> bool:
        """Current hold status. Hold enables an additional user input to
        enable and disable the automatic run of individual solvers. The hold is
        stored as an integer enabling nested holds throughout the system.

        See Also
        --------
        Solver.raise_hold
        Solver.lower_hold
        """
        return bool(self._hold)

    @hold.setter
    def hold(self, val: (bool, int)) -> None:
        self._hold = int(val)

    def raise_hold(self) -> None:
        """Raise the hold level by one."""
        self._hold += 1

    def lower_hold(self) -> None:
        """Lower the hold level by one."""
        self._hold += 1

    def force_solve(self) -> None:
        """Run the solver, regardless of the current solver state."""
        self._solve()
        # ToDo: Add timing and logging.

    def solve(self) -> None:
        """Run the solver if necessary and allowed.

        The solver will run if the solver is not already solved, the
        solver is not currently already running, and the solver does not
        currently have an enabled hold condition.
        """
        if not self.solved_calculating and not self.hold:
            self.force_solve()
    # endregion

    # region Solver Functions
    def _pre_solve(self) -> None:
        """Conduct standardised pre-run of the solver."""
        self._calculating = True

    def _solve_function(self) -> None:
        """This is the actual solve function of the solver."""
        raise NotImplementedError

    def _post_solve(self) -> None:
        """Conduct standardised post-run of the solver."""
        self._calculating = False
        self._recalculate = False

    def _solve(self) -> None:
        """Private solve function combining the pre-, core-, and post-solve
        functions."""
        self._pre_solve()
        self._solve_function()
        self._post_solve()
    # endregion

    # region Input Properties
    @property
    def tolerance(self) -> float:
        """Solver tolerance."""
        return self._tolerance

    @tolerance.setter
    def tolerance(self, val: float) -> None:
        if val < self._tolerance:
            self.reset()
        self._tolerance = val
    # endregion



from __future__ import annotations
from cetpy.Modules.SysML._ValueProperty import ValueProperty, value_property, \
    DeterminationTest, AggregateValueProperty, ValuePrinter
from cetpy.Modules.SysML._PartProperty import PartProperty
from cetpy.Modules.SysML._FlowProperty import FlowProperty
from cetpy.Modules.SysML._Block import Block
from cetpy.Modules.SysML._Port import Port
from cetpy.Modules.SysML._ContinuousPort import ContinuousPort



"""
FlowProperty
============

This file implements a basic Flow Property. A flow property acts like a
value property on a continuous flow port adding relevant solve functions to
solve the changes in the property along the continuous flow as well as deal
with over- and under determination, solver directions, and discontinuities.

An example of the use of the FlowProperty may be the pressure, temperature,
and mass-flow which could define the fluid condition in a fluid system. Note
the fluid itself, which describes the behaviour of the flow, should be a
flow item, shared by all fluid elements, not a FlowProperty. The chemical
composition of a reacting gas on the other hand, should be a FlowProperty.
"""

from __future__ import annotations

from typing import Any, Tuple, Callable, List

from cetpy.Modules.SysML._ValueProperty import ValueProperty, ValuePropertyDoc
from cetpy.Modules.Utilities.InputValidation import validate_input


class FlowProperty(ValueProperty):
    """FlowProperty which extends the ValueProperty to add solver functions
    along a continuous flow. Such as for pressure, temperature,
    and mass-flow in a fluid system."""
    __slots__ = ['_name_instance_input', '_name_instance_direction',
                 '_name_instance_recalculate']

    def __init__(self,
                 unit: str = None, axis_label: str = None,
                 necessity_test: float = 0.1) -> None:
        self._name_instance_input = ''
        self._name_instance_direction = ''
        self._name_instance_recalculate = ''
        super().__init__(unit, axis_label, necessity_test=necessity_test)

    # region Decorators
    def getter(self, fget: Callable) -> ValueProperty:
        raise AttributeError("Flow properties do not have getter methods.")

    def setter(self, fset: Callable) -> ValueProperty:
        raise AttributeError("Flow properties do not have setter methods.")

    def deleter(self, fdel: Callable) -> ValueProperty:
        raise AttributeError("Flow properties do not have deleter methods.")
    # endregion

    # region Getters and Setters
    def __set_name__(self, cls, name):
        ValueProperty.__set_name__(self, cls, name)
        self._name_instance_input = '_' + name + '_input'
        self._name_instance_direction = '_' + name + '_direction'
        self._name_instance_recalculate = '_d' + name + '_recalculate'
        cls.__flow_properties__ += [self]

    def __init_private_attributes__(self, instance):
        """Create relevant private attributes on the port."""
        instance.__setattr__(self._name_instance, 0)
        instance.__setattr__(self._name_instance_reset, 0)
        instance.__setattr__(self._name_instance_input, None)
        instance.__setattr__(self._name_instance_direction, 'downstream')

    def __create_block_delta_attribute__(self, instance) -> None:
        """Create an inlet and outlet delta attribute for port parent block."""
        name = self._name
        name_delta = 'd' + name
        name_update = '_d' + name + '_solve'
        name_recalculate = self._name_instance_recalculate
        name_stored = '_d' + name + '_stored'

        # Try if it exists, if yes, exit
        try:
            instance.__getattribute__(name_recalculate)
            return
        except AttributeError:
            pass  # Create attributes

        instance.__setattr__(name_recalculate, False)
        instance.__setattr__(name_stored, 0)
        reset_dict = type(instance).__getattribute__('_reset_dict')
        reset_dict.update({self._name_instance_recalculate: True})
        hard_reset_dict = type(instance).__getattribute__('_hard_reset_dict')
        hard_reset_dict.update({name_stored: 0})

        def delta() -> float:
            fr"""Difference in {name} from inlet to outlet [{self.unit}].
            
            ..math::
                \Delta {name} = {name}_o_u_t - {name}_i_n
            """
            if not instance.__getattribute__(name_recalculate):
                instance.__setattr__(name_recalculate, True)
                try:
                    instance.__setattr__(
                        name_stored, instance.__getattribute__(name_update))
                except AttributeError:
                    pass
                instance.__setattr__(name_recalculate, False)
            return instance.__getattribute__(name_stored)

        instance.__setattr__(name_delta, property(delta))

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return instance.__getattribute__(self._name_instance)

    def __set_converging_value__(self, instance, value) -> None:
        instance.__setattr__(self._name_instance, value)

    def __set__(self, instance, value) -> None:
        if not instance.is_endpoint:  # Verify instance is a valid input point
            raise ValueError("A boundary condition cannot be set on an "
                             "internal point of a flow system.")

        val_initial_inp = instance.__getattribute__(self._name_instance_input)
        val_initial = instance.__getattribute__(self._name_instance_reset)

        value = validate_input(value, self.permissible_types_list,
                               self.permissible_list, self._name)
        self.__set_converging_value__(instance, value)
        instance.__setattr__(self._name_instance_input, value)

        if val_initial_inp is None and value is not None:
            # Port was not previously a boundary condition -> update solver
            # direction.
            port_list = instance.flow_system_port_list

            # Evaluate direction and remove opposite boundary condition.
            if instance.is_downstream_endpoint:
                direction = 'upstream'
                port_list[0].__setattr__(self._name_instance_input, None)
            else:
                direction = 'downstream'
                port_list[-1].__setattr__(self._name_instance_input, None)
            for p in port_list:
                self.set_direction(p, direction)

        if self._determination_test is not None:
            self._determination_test.test(instance, self._name)

        if self.__reset_necessary__(instance, value, val_initial):
            instance.__setattr__(self._name_instance_reset, value)
            instance.reset()

    def __delete__(self, instance):
        delattr(instance, self._name_instance)
    # endregion

    # region Input properties
    def get_direction(self, instance) -> str:
        """Solve direction of the flow property at the instance. 'downstream'
        means that the property is set upstream and solved downstream.
        'upstream' means the opposite."""
        return instance.__getattribute__(self._name_instance_direction)

    def set_direction(self, instance, val: str) -> None:
        """Set the solve direction at the instance."""
        if val not in ['upstream', 'downstream']:
            raise ValueError("Direction must be either up- or downstream.")
        for p in instance.flow_system_port_list:
            p.__setattr__(self._name_instance_direction, val)
    # endregion



"""
Congruent Engineering Toolbox (CET) Session
===========================================

This file defines a CET program session, which defines the configuration,
program parameters, and working directory.
"""

import logging
from importlib import reload
from os.path import join, split, dirname
import sys

import cetpy.Configuration
from cetpy.Configuration import ConfigurationManager


class Session:
    """Congruent Engineering Toolbox Session."""

    def __init__(self, directory: str | None, logging_level: str = 'info'):
        self.logger = None
        self.config_manager = ConfigurationManager(directory)
        self.logging_level = logging_level
        self.directory = directory

    def parameter(self, name: str):
        """Return key from the session config dictionary.

        Parameters
        ----------
        name
            The name of the property value being requested.

        Raises
        ------
        KeyError: If the name is not in the dictionary.
        """
        return self.config_manager.parameter(name)

    @property
    def config_dict(self) -> dict:
        """Session config dictionary defining system and program parameters."""
        return self.config_manager.config

    @property
    def name(self) -> str:
        """Return session name."""
        return split(self.directory)[-1]

    @property
    def directory(self) -> str:
        """Session working directory, for session configs and saving."""
        return self._directory

    @directory.setter
    def directory(self, val: str) -> None:
        self._directory = val
        self.config_manager.directory = val
        self.start_logging()

    def refresh(self) -> None:
        """Refresh this session. Does not restart logging."""
        self.config_manager.reset()
        self.config_manager.load()
        cetpy.Configuration.refresh_module_dict()

    def restart(self) -> None:
        """Restart this session. This both refreshes the config files and
        restarts the logging."""
        self.shutdown()
        self.refresh()

    @staticmethod
    def shutdown() -> None:
        """Shutdown session and logging."""
        logging.shutdown()

    def start_logging(self) -> None:
        """Start a logging process for this session."""
        reload(logging)

        match self.logging_level:
            case 'debug':
                log_level = logging.DEBUG
            case 'info':
                log_level = 15
            case 'warning':
                log_level = logging.WARNING
            case 'critical':
                log_level = logging.CRITICAL
            case _:
                log_level = self.logging_level

        # Log File
        formatter = logging.Formatter(
            '[%(asctime)s:%(filename)15s:%(funcName)25s:%(lineno)3d:'
            '%(levelname)8s]: %(message)s')

        directory = self.directory
        if directory is None:
            directory = dirname(__file__)
            name = 'Base'
        else:
            name = self.name

        handler = logging.FileHandler(join(directory, 'cetpy.log'),
                                      mode='w')
        handler.setFormatter(formatter)
        handler.setLevel(log_level)

        # Console
        handler_console = logging.StreamHandler()
        handler_console.setStream(sys.stdout)
        handler_console.setLevel(logging.INFO)

        logger = logging.getLogger(name)
        logger.setLevel(log_level)
        logger.addHandler(handler)
        logger.addHandler(handler_console)
        logging.captureWarnings(False)
        self.logger = logger



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name='cetpy',
    version='0.1.0b1',
    description='Congruent Engineering Toolbox',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/CongruentEngineeringToolbox/cetpy",
    author='CET developers',
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords="engineering, system engineering, congruent engineering",
    packages=find_packages(include=['cetpy', 'cetpy.*']),
    include_package_data=True,
    python_requires=">=3.10, <4",
    install_requires=[
        'numpy',
        'pandas',
        'CoolProp',
        'tomli',
    ],
)



from os.path import dirname, join

from cetpy.Configuration._ConfigurationManager import ConfigurationManager
from cetpy.Configuration._Session import Session
from cetpy.Configuration._ModuleManager import \
    generate_block_class_list, get_module

config_locations = [join(dirname(__file__), 'Configurations')]
module_locations = [join(dirname(__file__), '..', 'Modules')]
module_dict = {}


def refresh_module_dict() -> None:
    """Refresh the list of modules of cetpy and extension modules."""
    module_dict.clear()
    module_dict.update(generate_block_class_list())
